/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package content.exercises;

import matrix.animation.Animator;
import matrix.util.Note;

public class MST_Prim extends matrix.structures.FDT.probe.CommonLabelTreeImpl {
    PriorityExerVertex[] g;
    PriorityExerVertex r;
	
    public MST_Prim(String label) {
	super(label);
	setLabelVisible(false);
	setLabel("0");
	setReferenceLabelVisible(false);
    }
	
    public MST_Prim(PriorityExerVertex[] g, PriorityExerVertex r) {
	this("");
	this.g=g;
	this.r=r;
	
    }

    class PriorityQueue {
	PriorityExerVertex[] array;
	int size = 0;
	int maxsize; 

	//debug
	void printQ() {
	    String s = " ";
	    for (int i=0; i<size; i++)
		s+="["+i+"]"+array[i]+"("+((array[i] != null) ? ""+array[i].priority : "*")+")";
	    Note.out(this, s);
	}

	public PriorityQueue() {
	    array = new PriorityExerVertex[maxsize=8];
	}
	    
	public boolean isEmpty() {
	    return (size == 0);
	}

	public void insert(PriorityExerVertex v) {
	    //Note.out(this, "Inserting " + v + " with priority "+v.getPriority());
	    //Note.out(this, " -BEFORE-");
	    //printQ();
	    if (size == maxsize) 
		array = duplicateHeapSize(array);
	    array[size] = v;
	    size=size+1;
	    heapify(size-1);
	    //Note.out(this, " -AFTER-");
	    //printQ();
	    //Note.out(this, "Insert DONE.");
	}
	    
	// Perfomance issue: this operation requires O(N) time even though
	// it is possible to implement a version that requires only O(log N) time.
	// An additional data structure would be needed for that.
	// However, as the heap size is very small here, we prefer to make this
	// simpler and not use any additional data structure to make this faster.
	public void update(PriorityExerVertex v) {
	    //Note.out(this, "Updating " + v + " with priority "+v.getPriority());
	    //Note.out(this, " -BEFORE-");
	    //printQ();
	    for (int i=0; i<size; i++)
		if (array[i] == v) {
		    heapify(i);
		    //Note.out(this, " -AFTER-");
		    //printQ();
		    //Note.out(this, "Insert DONE.");
		    return;
		}
	    Note.warning(this, "Trying to update an item that doesn't exists in the heap.");
	}

	public PriorityExerVertex deleteMin() {
	    //Note.out(this, "Deleting " + array[0]);
	    if (size == 0) {
		Note.warning(this, "Attempt to delete min from empty heap");
		return null;
	    }
	    PriorityExerVertex tmp = array[0];
	    size=size-1;
	    array[0]=array[size];
	    percolateDown(0);
	    //printQ();
	    return tmp;
	}
	    
	// Return 0 if priority of item at array[i] is less than the priority of the item at array[j]
	// Return 1 if it's the other way around
	// In case the priority is equal the order is determined by looking at the time stamp
	// The oldest item has then the highest priority (should removed first)
	private int cmpPriority(int i, int j) {
	    if (array[i].getPriority() < array[j].getPriority())
		return 0;
	    if (array[i].getPriority() > array[j].getPriority())
		return 1;
	    if (array[i].getTimeStamp() <= array[j].getTimeStamp())
		return 0;
	    return 1;
	}

	private void heapify(int i) {
	    //Note.out(this, "Heapify  "+i);
	    //Note.out(this, "array[i]="+array[i]);
	    if (cmpPriority(father(i), i) == 0) {
		//Note.out(this, "ready: ");
		//printArray();
		return;
	    }
	    //Note.out(this, "SWAP");
	    swap(father(i), i);
	    //Note.out(this, "array[i]="+array[i]);
	    heapify(father(i));
	}
	    
	private void percolateDown(int i) {
	    //Note.out(this, "PercolateDown i="+i + " with heap size " + size);
	    //Note.out(this, "left = "+left(i)+"; right="+right(i));
	    //Note.out(this, "array[i]="+array[i]);
	    if (isEmpty())
		return;
	    int j;
	    if (cmpPriority(left(i), right(i)) == 0) {
		//Note.out(this, "left smaller " + left(i));
		if (cmpPriority(i, j = left(i)) == 0)
		    return;
	    } else {
		//Note.out(this, "right smaller " + right(i));
		if (cmpPriority(i, j = right(i)) == 0)
		    return;
	    }
	    swap(i,j);
	    percolateDown(j);
	}

	private void swap(int i, int j) {
	    //Note.out(this, "Swap " + i + " and " + j);
	    PriorityExerVertex tmp = array[i];
	    array[i] = array[j];
	    array[j] = tmp;
	}

	private int father(int i) {
	    if (i==0)
		return 0;
	    else
		return (i-1)/2;
	}

	private int left(int i) {
	    int n = 2*i+1;
	    if (n >= size)
		return i;
	    else
		return n;
	}

	private int right(int i) {
	    int n = 2*i+2;
	    if (n >= size)
		return i;
	    else
		return n;
	}

	private PriorityExerVertex[] duplicateHeapSize(PriorityExerVertex[] array) {
	    //Note.out(this, "Duplicate heap");
	    PriorityExerVertex[] newarray = new PriorityExerVertex[maxsize=2*maxsize];
	    for (int i=0; i<array.length; i++)
		newarray[i]=array[i];
	    return newarray;
	}
    }

    public Object getElement() {
	return r;
    }
	
    public PriorityExerVertex run() {
	Animator a = Animator.getActiveAnimator();
	PriorityQueue Q = new PriorityQueue();
	//Note.out(this, "Running MST_Prim...");
	for (int u=0; u<g.length; u++) {
	    g[u].setPriority(Integer.MAX_VALUE);
	    g[u].setVisited(false);
	    g[u].setFather(null);
	}
	a.startOperation();
	//Note.out(this, "Init complete");
	r.setPriority(0);
	//Note.out(this, "Inserting root into Q...");
	Q.insert(r);
	int count = 0;
	while (!Q.isEmpty()) {
	    PriorityExerVertex u;
	    u = Q.deleteMin();
	    if (u.getVisited())
		continue;
	    //Note.out(this, "->u="+u);
	    u.setVisited(true);
	    u.setLabel(++count+"");
	    // add edge into the spanning tree
	    //Note.out(this, "add edge " + u.getFather() + "->" + u);
	    PriorityExerVertex f = u.getFather();
	    if (f != null) {
		f.referenceSelectedTo(u);
		a.endOperation();
		a.startOperation();
	    }
	    PriorityExerVertex[] pev = u.getAdjacentVertices();
	    for (int i=0; i<pev.length; i++) {
		PriorityExerVertex v = pev[i];
		int w = u.getWeight(v);
		if ((!v.getVisited()) && (w < v.getPriority())) {
		    //Note.out(this, "Q.insert " + v);
		    v.setFather(u);
		    if (v.getPriority() < Integer.MAX_VALUE) {
			v.setPriority(w);
			Q.update(v);
		    } else {
			v.setPriority(w);
			Q.insert(v);
		    }
		}
	    }
	}
	a.endOperation();
	return r;
    }
}

