/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: RadixExchangeSort.java,v 1.18 2004/05/26 08:05:45 vkaravir Exp $ */

package content.exercises;

import java.awt.*;
import java.awt.event.*;
import java.util.Random;

import matrix.animation.Animator;
import matrix.simulation.VisualTypeConf;
import matrix.structures.CDT.probe.Stack;
import matrix.structures.EDT.ColorableKey;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.Table;
import matrix.structures.memory.*;
import matrix.util.*;
import matrix.visual.VisualType;


public class RadixExchangeSort implements SimulationExerciseModel,
					  StyledExercise, ModelAnswerNames, 
					  ConfigureVisualType, AWTComponentUtilizer,
					  ButtonGroups {

    private Application app;

    public void setApplication(Application a) {
	app=a;
    }

    public Application getApplication() {
	return app;
    }

    /* how many odd numbers are required in the array to be sorted */
    public static final int ODD_NUMBER_COUNT = 5;

    public static final int ARRAY_SIZE = 10;

    /* indices of the left and right border in the parameter table */
    public static final int LEFT_BORDER = 0;
    public static final int RIGHT_BORDER = 1;


    /* length of keys */
    public static final int KEY_BITS = 3;


    private Table userTable;
    private BitSelectionTable userBitSelect;
    private Stack userStack;

    protected String[] keyArray = null;
    public static final boolean DEBUG = false;

    long seed = 1;

    /**
     * Get the value of seed.
     * @return value of seed.
     */
    public long getSeed() {
        return seed;
    }

    /**
     * Set the value of seed.
     * @param v  Value to assign to seed.
     */
    public void setSeed(long  v) {
        this.seed = v;
    }

    public FDT[] init() {
	java.util.Random r =  new java.util.Random(seed);
	BitTable bitTable = null;
	do {
	    bitTable = createBitTable(r, 0, 25, KEY_BITS, ARRAY_SIZE);
	} while (bitTable.oddCount < ODD_NUMBER_COUNT);
	keyArray = bitTable.table;

	userTable = new Table(makeKeys(keyArray));

	SelectableKey[] bitKeys = new SelectableKey[KEY_BITS];
	for(int i=0; i<bitKeys.length; i++)
	    bitKeys[i]=new SelectableKey(bitKeys.length -1 - i);

	userBitSelect = new BitSelectionTable(bitKeys);

	userStack = new Stack();
	FDT[]  s = { userTable, userBitSelect, userStack };
	return s;
    }

    public String[] getStructureNames() {
	String[] s = { "Table of Keys to be sorted", "Current Bit", "Stack"};
	return s;
    }

    public String[] getStructureVisualisations() {
	String[] s = { "array", "array", "list" };
	return s;
    }


    public String[] getModelAnswerVisualisations() {
	String[] s = { "array", "list" };
	return s;
    }


    public String getDescription() {
	return "Just sort it!";
    }


    public String[] getModelAnswerNames() {
	String[] s = {  "Table of Keys to be sorted", "Stack"};
	return s;
    }


    public FDT[] getAnswer() {
	FDT[] s = { userTable, userStack };
	return s;
    }


    public FDT[] init(Object input) {
	return init();
    }


    public FDT[] solve() {
	Table mTable = (Table)getInitialStructures()[0];
	Table mBorders = (Table)getInitialStructures()[1];
	Stack mStack = (Stack)getInitialStructures()[2];
	Animator a = Animator.getActiveAnimator();
	radexchSort(a, mTable, mStack,  false);
	return new FDT[] {mTable, mStack};
    }


   
    public FDT[] makeModelAnswer() {
	Table mTable = (Table)getInitialStructures()[0];
	Table mBorders = (Table)getInitialStructures()[1];
	Stack mStack = (Stack)getInitialStructures()[2];
	Animator a = Animator.getActiveAnimator();
	radexchSort(a, mTable, mStack, true);
	return new FDT[] {mTable, mStack};
    }



    /**
       Sorts the given table using iterative radix exchange sort.

       @param a animator for animating the algorithm
       @param table the table to be sorted
       @param size the size of the table
       @param stack the recursion stack
       @param highlightOn a flag indicating whether to highlight keys in the table when
       the algorithm proceeds. Highlighting is in generating the model solution.
    */
    static void radexchSort(Animator a, Table table, 
			     Stack stack, boolean highlightOn) {
	int size = table.size();
	int bottom = 0;
	int top = size-1;
	int bits = KEY_BITS-1;

	a.startOperation();
	pushParameters(stack, new int[]{bottom, top, bits});
	a.endOperation();

	int[] parameters = null;
	while ((parameters = popParametersAnimated(stack, a)) != null) {
	    bottom = parameters[0];
	    top = parameters[1];
	    bits = parameters[2];
	    partition(a, stack, table, bottom, top, bits, highlightOn);
	}
    }	
    


    static int[] popParametersAnimated(Stack stack, Animator a) {
	a.startOperation();
	int [] arr=popParameters(stack);
	a.endOperation();
	return arr;
    }



    static void partition(Animator a, Stack stack, Table table, int bottom,
			  int top, int b, boolean highlightOn) {
	if (top > bottom && b >= 0) {
	    int lo = bottom, hi = top;

	    if (highlightOn) {
		a.startOperation();
		clearHighlight(table);
		highlightTable(table, lo, hi, Color.yellow);
		a.endOperation();
	    }
	    
	    while (lo != hi) {
		while (bit(table.getObject(lo).toString(), b)==0 && lo<hi) {
		    lo++;
		}
		
		while (bit(table.getObject(hi).toString(), b)!=0 && lo<hi) {
		    hi--;
		}

		if (lo != hi) {
		    a.startOperation();
		    Object lowKey = table.getObject(lo);
		    Object highKey = table.getObject(hi);
		
		    if (highlightOn) {
			// highlight keys

			// TODO: lowKey.setSwapped(true); hihgKey.setSwapped(true);
			highlightKey(lowKey, Color.green);
			highlightKey(highKey, Color.green);
			a.endOperation();
			a.startOperation();
		    }

		    table.setObject(highKey, lo);
		    table.setObject(lowKey, hi);
		    a.endOperation();	
		    
		    // remove highlighting
		    if (highlightOn) {
			a.startOperation();
			// TODO: lowKey.setSwapped(false); hihgKey.setSwapped(false);
			highlightKey(lowKey, Color.yellow);
			highlightKey(highKey, Color.yellow);
			a.endOperation();
		    }
		}
	    }
	    
	    if (bit(table.getObject(top).toString(), b)==0)
		hi++;
	   
	    // push borders to stack
	    if (b>0) {
		a.startOperation();
		pushParameters(stack, new int[]{bottom, hi-1, b-1});
		pushParameters(stack, new int[]{hi, top, b-1});
		a.endOperation();
	    }
	}
    }	



    static void highlightTable(Table t, int lo, int hi, Color color) {
	for (int i = lo; i <= hi; i++) {
	    Object o = t.getObject(i);
	    if (o instanceof ColorableKey) {
		highlightKey(((ColorableKey)o), color);
	    }
	}
    }
    

    static void highlightKey(Object k, Color c) {
	if (k instanceof ColorableKey)
	    ((ColorableKey)k).getStyleSheet().setDefaultBackgroundColor(c);
    }


    
    static void clearHighlight(Table t) {
	for (int i = 0; i < t.size(); i++) {
	    Object o = t.getObject(i);
	    if (o instanceof ColorableKey) {
		((ColorableKey)o).setDefaultBackgroundColor(Color.white);
	    }
	}
    }
    


    static void pushParameters(Stack stack, int[] params) {
	Table t = new Table(params.length);
	setParameters(t, params);
	stack.insert(t);
    }
    
    
    static void setParameters(Table t, int[] params) {
	for (int i = 0; i < params.length; i++) {
	    t.setObject(params[i]+"", i);
	}
    }


    static int[] popParameters(Stack stack) {
	Table params = (Table)stack.getTopOfStack();
	if (params==null)
	    return null;
	stack.pop(null);
	return getParameters(params);
    }
    

    static int bit(int x, int which) {
	return ((x >> which) & 1);
    }

    
    static int bit(String s, int which) {
	return s.charAt(s.length()-which-1)=='1'?1:0;
    }


    public FDT[] getInitialStructures() {
	return new FDT[] { new Table(makeKeys(keyArray)), new Table(new String[] {" ", " "}), 
			   new Stack()};
    }


    public FDT[] getSimulatedStructures() {
	return new FDT[] { userTable, userStack };
    }


    public SimulationExerciseModel getModelAnswer() {
	return this;
    }


    public VisualTypeConf[] conf() {
	VisualTypeConf tabl = new VisualTypeConf();
	tabl.enable("matrix.visual.VisualKey", VisualTypeConf.HIGHLIGHT_OPERATION);
	tabl.enable("matrix.visual.VisualKey", VisualTypeConf.DRAG_OPERATION);
	tabl.enable("matrix.visual.VisualKey", VisualTypeConf.DROP_OPERATION);
	tabl.enable("matrix.visual.VisualArrayComponent", VisualTypeConf.HIGHLIGHT_OPERATION);

	VisualTypeConf bitSelect = new VisualTypeConf();
	bitSelect.enable("matrix.visual.VisualKey", VisualTypeConf.HIGHLIGHT_OPERATION);

	VisualTypeConf stack = new VisualTypeConf();
	return new VisualTypeConf[] { tabl, bitSelect, stack };
    }



    /**
       Creates a bit array.
       
       @param r random number generator
       @param bits number of bits used
       @param size number of characters of the key to create
    */
    public BitTable createBitTable(Random r, int min, int max, int bits,
					  int size) {
	String[] str = new String[size];
	int i = 0;
	int oddCount = 0;
	while (i < size) {
	    // Ascii codes of letters A-Z are from 1 to 26.
	    int rand = ((int)(r.nextDouble() * 25))+1;
	    str[i] = binaryRepresentation(rand, bits);
	    if (rand % 2 == 1) oddCount++;
	    i++;
	}
	return new BitTable(str, oddCount);
    }
    
    
   
    public static String binaryRepresentation(int n, int bits) {
	StringBuffer buf = new StringBuffer();
	for (int i = bits-1; i >= 0; i--) {
	    buf.append(((n >> i) % 2) == 0 ? "0":"1");
	}
	return buf.toString();
    }



    // Buttons

    public Component[] getAWTComponents() {
/*	Button setLeft = new Button("Set left");
	setLeft.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent evt) {
		    setBorderParameter(LEFT_BORDER);
		}
	    });
	
	Button setRight = new Button("Set right");
	setRight.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent evt) {
		    setBorderParameter(RIGHT_BORDER);
		}
	    });
*/
	Button push = new Button("Push");
	push.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent evt) {
		    pushButton();
		}
	    });
	
	Button pop = new Button("Pop");
	pop.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent evt) {
		    popButton();
		}
	    });

/*	Button[] bitButtons = createBitButtons(KEY_BITS);

	Button[] allButtons = new Button[bitButtons.length+4];

	allButtons[0]=setLeft;
	allButtons[1]=setRight;
	allButtons[2+bitButtons.length]=push;
	allButtons[3+bitButtons.length]=pop;
	System.arraycopy(bitButtons, 0, allButtons, 2, bitButtons.length);*/
	return new Component[]{push, pop};
    }

/*
    Button[] createBitButtons(int bits) {
	Button[] b = new Button[bits];
	for (int i = 0; i < bits; i++) {
	    final int bit=(bits-i-1);
	    b[i] = new Button("Bit" + bit);

	    b[i].addActionListener(new ActionListener() {
		    public void actionPerformed(ActionEvent evt) {
			setBit(bit);			
		    }
		});

	    
	}
	return b;
    }
    


    void setBorderParameter(int parameterIndex) {
	if (VisualType.read == null || 
	    !(VisualType.read instanceof VisualArrayComponent) ||
	    !containsStructure(VisualType.read, userTable)) {
	    Note.show(null, "Select an index from the table.");
	    return;
	}
	
	VisualArrayComponent a = 
	    (VisualArrayComponent)VisualType.read;
	int index = a.getIndex();
	userParameters.setObject(index+"", parameterIndex);
	validateApplication();
    }
  

    
    void setBit(int bit) {
	userParameters.setObject(bit+"", 2);
	validateApplication();
    }
*/


    boolean containsStructure(VisualType vt, FDT s) {
	if (vt == null)
	    return false;
	else if (vt.getStructure() == null)
	    return false;
	else if (vt.getStructure() == s) 
	    return true;
	else if (!(vt.getParent() instanceof VisualType))
	    return false;
	else
	    return containsStructure((VisualType)vt.getParent(), s);
    }




    static int[] getParameters(Table params) {
	int[] p = new int[params.size()];
	for (int i = 0; i < p.length; i++) {
	    try {
		String s=params.getObject(i).toString();
		p[i] = Integer.valueOf(s).intValue();
	    } catch (NumberFormatException e) {		
		p[i] = -1;
	    } catch (NullPointerException e) {
		p[i] = -1;
	    }
	}
	return p;
    }


    void pushButton() {
	int[] params =new int[3]; 

	params[0] = userBitSelect.getSelectedKeyValue();
	params[1] = userTable.selectionStart();
	params[2] = userTable.selectionEnd();

/*getParameters(userParameters);
	for (int i = 0; i < params.length; i++) {
	    if (params[i] < 0) {
		Note.show(null, "Select parameters.");
		return;
	    }
	    }*/
	getApplication().getAnimator().startOperation();
	pushParameters(userStack, params);
	getApplication().getAnimator().endOperation();
	//clearTable(userParameters);
	
	validateApplication();
    }
    
    

    void clearTable(Table t) {
	for (int i = 0; i < t.size(); i++) {
	    t.setObject(" ", i);
	}
    }


    
    void popButton() {
	getApplication().getAnimator().startOperation();
	int[] parameters = popParameters(userStack);
	Animator.getActiveAnimator().endOperation();
	//setParameters(userParameters, parameters);
	
	userTable.selectRange(parameters[1], parameters[2]);
	userBitSelect.setSelectedKeyValue(parameters[0]);

	validateApplication();
    }
    


    // Buttongroup

    public int[] buttonGroups() {
        int[] n = new int[3];
        n[0] = 0;
        n[1] = 2;
        n[2] = 2+KEY_BITS;
        return n;
    }




    void validateApplication() {
	if (app != null) {
	    app.validate();
	    app.validateAnimator();
	}
    }


    
    private ColorableKey[] makeKeys(String[] keys) {
	ColorableKey[] vKeys = new ColorableKey[keys.length];
	for (int i = 0; i < keys.length; i++) {
	    vKeys[i] = new ColorableKey(keys[i]);
	} 
	return vKeys;
    }


    
    
    class BitTable {
	String[] table;
	int oddCount;
	BitTable(String[] table_, int oddCount_) {
	    table = table_;
	    oddCount = oddCount_;
	}
    }


    class BitSelectionTable extends Table implements StyledArray {
	SelectableKey current;

	BitSelectionTable(SelectableKey[] args) {
	    super(args);

	    for (int i=0; i<args.length; i++){
		args[i].setParent(this);
	    }

	    current = args[0];
	    current.set();

	    
	}

	public String getIndexName(int address) {
	    return "";	    
	}

	public void setSelectedKey(SelectableKey caller){
	    current.clear();
	    caller.set();
	    current = caller;
	}

	public int getSelectedKeyValue(){
	    return current.getValue();
	}

	public void setSelectedKeyValue(int index){
	    Note.out(this, "set selected " + index);
	    Note.out(this, "that is " + getObject(index));
	    setSelectedKey((SelectableKey)getObject(KEY_BITS-1-index));
	}
    }

    class SelectableKey extends Key{
	BitSelectionTable parent;
	int keyValue;
	VirtualBoolean isSelected = new VirtualBoolean(false);

	public boolean isSelected(){
	    Note.out(this, "focusrequest "+isSelected.eval());
	    return isSelected.eval();
	}

	public void set(){
	    Note.out(this, "set "+keyValue);
	    isSelected.assign(true);
	}

	public void clear(){
	    isSelected.assign(false);
	}

	public SelectableKey(int keyValue){
	    super("Bit "+keyValue);
	    this.keyValue = keyValue;
	}

	public void setParent(BitSelectionTable parent){
	    this.parent = parent;
	}

	public void select(){
	    //Note.out(this, "select Entity "+this);
	    parent.setSelectedKey(this);
	}	

	public int getValue(){
	    return keyValue;
	}
    }

    public static void main(String[] args) {
	for (int i = 1; i <= 16; i++) {
	    System.out.println(binaryRepresentation(i,5));
	}
    }
}
