/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: StyleSheetAdapter.java,v 1.9 2004/05/26 08:05:49 vkaravir Exp $ */

package matrix.decoration;

import java.awt.Color;

import matrix.structures.FDT.FDT;
//import matrix.visual.LabelDecorator;
//import matrix.visual.ColorDecorator;

/**
 * Default implemetation of StyleSheet decorator. Classes that require some styling
 * functions should override the methods of this class as required.
 * @see StyleSheet
 * @see ColorDecorator
 * @see LabelDecorator
 */

public class StyleSheetAdapter implements StyleSheet {
    private static StyleSheetAdapter singleton = null;

    /** Default background color; white. */
    private Color defaultBackgroundColor = Color.white;

    /** Default background color for visited elements; green. */
    private Color visitedBackgroundColor = Color.green;

    /** Default background color for focused elements; blue. */
    private Color focusedBackgroundColor = Color.blue;

    /** Default background color for visited, focused elements; magenta. */
    private Color visitedFocusedBackgroundColor = Color.magenta;

    /** Default border color; blue. */
    private Color defaultBorderColor = Color.blue;

    /** Default border color for visited elements; green. */
    private Color visitedBorderColor = Color.yellow;

    /** Default border color for focused elements; blue. */
    private Color focusedBorderColor = Color.blue;

    /** Default border color for visited, focused elements; green. */
    private Color visitedFocusedBorderColor = Color.green;

    /** Default pen color; red. */
    private Color defaultPenColor = Color.red;

    /** Default pen color for visited elements; red. */
    private Color visitedPenColor = Color.red;

    /** Default pen color for focused elements; white. */
    private Color focusedPenColor = Color.white;

    /** Default pen color for visited, focused elements; white. */
    private Color visitedFocusedPenColor = Color.white;

  /** Default color for references; blue */
  private Color referenceColor = Color.blue;

  /** Color for focused references; red */
  private Color focusedReferenceColor = Color.red;

  /** Color for modified references; green */
  private Color modifiedReferenceColor = Color.red;

    private String extraInfo = "no additional information available";

    private int infoWaitTime = 2000;

    public StyleSheetAdapter() {}

    /* This is the static singleton method to get instance from this class.
    public static StyleSheetAdapter getStyleSheetAdapter() {
        if (singleton == null)
            return (singleton = new StyleSheetAdapter());
        else
            return singleton;
    }
    */

    // --- implements StyleSheet ---

    public StyleSheet getStyleSheet() {
        return this;
    }

    //  --- implements InformationDecorator ---

    public boolean isInfoEnabled() { return false; }

    public void setInfo(String info) { extraInfo = info; }

    public String getInfo() { return extraInfo; }

    public int getWaitTime() { return infoWaitTime; }

    public void setWaitTime(int t) { if (t>=0) infoWaitTime = t; }

    // --- implements LabelDecorator ---

    /**
     * Tells if labeling is enabled for nodes.
     * By default this adapter return false.
     * @return true if label is shown
     */
    public boolean isLabelEnabled() { return false; }

    /**
     * Sets the label.
     * @param s the new label
     */
    public void setLabel(String s) {}

    /**
     * Returns the label.
     * @return the label
     */
    public String getLabel() { return "label"; }

    /**
     * Tells if labeling is enabled for references.
     * By default this adapter return false.
     * @return true if label is shown
     */
    public boolean isReferenceLabelEnabled() { return false; }

    /**
     * Sets the label of the i'th reference.
     * @param i reference number
     * @param s the new label
     */
    public void setReferenceLabel(int i, String s) {}

    /**
     * Returns the i'th label.
     * @param i the reference number
     * @return the label
     */
    public String getReferenceLabel(int i) { return "label"; }

    // --- implements ColorDecorator ---

    // Colors that paint the background (ie what is not Border or Pen)
    // of the element.

    /** Sets the default background color.
     * @param c the new color
     */
    public void setDefaultBackgroundColor(Color c) { defaultBackgroundColor = c; }
    /**
     * Sets the focused background color.
     * @param c the new color
     */
    public void setFocusedBackgroundColor(Color c) { focusedBackgroundColor = c; }
    /**
     * Sets the visited background color.
     * @param c the new color
     */
    public void setVisitedBackgroundColor(Color c) { visitedBackgroundColor = c; }

    /**
     * Sets the visited focused background color.
     * @param c the new color
     */
    public void setVisitedFocusedBackgroundColor(Color c) {visitedFocusedBackgroundColor = c; }

    /**
     * Gets the default background color.
     * @return the color
     */
    public Color getDefaultBackgroundColor() { return defaultBackgroundColor; }

    /**
     * Gets the focused background color.
     * @return the color
     */
    public Color getFocusedBackgroundColor() { return focusedBackgroundColor; }

    /**
     * Gets the visited background color.
     * @return the color
     */
    public Color getVisitedBackgroundColor() { return visitedBackgroundColor; }

    /**
     * Gets the visited focused background color.
     * @return the color
     */
    public Color getVisitedFocusedBackgroundColor() { return visitedFocusedBackgroundColor; }

    // Colors the paint the border of the element.

    /**
     * Sets the default border color
     * @param c the new color
     */
    public void setDefaultBorderColor(Color c) { defaultBorderColor = c;}

    /**
     * Sets the focused border color
     * @param c the new color
     */
    public void setFocusedBorderColor(Color c) { focusedBorderColor = c;}

    /**
     * Sets the visited border color
     * @param c the new color
     */
    public void setVisitedBorderColor(Color c) { visitedBorderColor = c;}

    /**
     * Sets the visited focused border color
     * @param c the new color
     */
    public void setVisitedFocusedBorderColor(Color c) { visitedFocusedBorderColor = c; }

    /**
     * gets the default border color
     * @return the color
     */
    public Color getDefaultBorderColor() { return defaultBorderColor; }

    /**
     * gets the focused border color
     * @return the color
     */
    public Color getFocusedBorderColor() { return focusedBorderColor; }

    /**
     * gets the visited border color
     * @return the color
     */
    public Color getVisitedBorderColor() { return visitedBorderColor; }

    /**
     * gets the visited focused border color
     * @return the color
     */
    public Color getVisitedFocusedBorderColor() { return visitedFocusedBorderColor; }

    // Colors that paint whatever is written or drawn inside the element
    // (ie keyvalue, title, etc)

    /**
     * Sets the default pen color.
     * @param c the new color
     */
    public void setDefaultPenColor(Color c) { defaultPenColor = c; }
    /**
     * Sets the focused pen color.
     * @param c the new color
     */
    public void setFocusedPenColor(Color c) { focusedPenColor = c;}
    /**
     * Sets the visited pen color.
     * @param c the new color
     */
    public void setVisitedPenColor(Color c) { visitedPenColor = c;}
    /**
     * Sets the visited focused pen color.
     * @param c the new color
     */
    public void setVisitedFocusedPenColor(Color c) { visitedFocusedPenColor = c; }

    /**
     * Gets the default pen color.
     * @return the color
     */
    public Color getDefaultPenColor() { return defaultPenColor; }

    /**
     * Gets the focused pen color.
     * @return the color
     */
    public Color getFocusedPenColor() { return focusedPenColor; }

    /**
     * Gets the visited pen color.
     * @return the color
     */
    public Color getVisitedPenColor() { return visitedPenColor; }

    /**
     * Gets the visited focused pen color.
     * @return the color
     */
    public Color getVisitedFocusedPenColor() { return visitedFocusedPenColor; }

  /**
   * Sets the color of a reference from
   * this to the parameter FDT.
   *
   * @param f the FDT the reference points to
   * @param c the new color of the reference
   */
  public void setReferenceColor(FDT f, Color c) {
    referenceColor = c;
  }

  /**
   * Sets the color of a focused reference.
   *
   * @param f the FDT the reference points to
   * @param c the new color
   */
  public void setFocusedReferenceColor(FDT f, Color c) {
    focusedReferenceColor = c;
  }


  /**
   * Sets the color of a modified reference
   *
   * @param f the FDT the reference points to
   * @param c the new color
   */
  public void setModifiedReferenceColor(FDT f, Color c) {
    modifiedReferenceColor = c;
  }

  /**
   * Gets the color of a reference from this to
   * the parameter FDT.
   *
   * @param f the destination of the reference
   * @return the color
   */
  public Color getReferenceColor(FDT f) {
    return referenceColor;
  }

  /**
   * Gets the color of focused reference
   *
   * @param f the FDT the reference points to
   * @return the color
   */
  public Color getFocusedReferenceColor(FDT f) {
    return focusedReferenceColor;
  }

  /**
   * gets the color of a modified reference
   *
   * @param f the FDT the reference points to
   * @return the color
   */
  public Color getModifiedReferenceColor(FDT f) {
    return modifiedReferenceColor;
  }



  static final long serialVersionUID = -9068980988985745464L;
}
