/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: SearchTable.java,v 1.3 2004/05/26 08:05:51 vkaravir Exp $ */

package matrix.structures.CDT.probe;

import matrix.animation.Animator;
import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.Table;
import matrix.structures.code.*;
import matrix.structures.memory.*;
import matrix.util.Note;


/**
 * Abstraction that takes an array (P) as it's input and outputs the start index or
 * indeces of all occurences of P in S by using the given method.
 * Example1: P could be an Element and the method could be linear search algorithm
 * Example2: P could be an Array and the method could be Boyer-Moore-Horspool algorithm
 * The subclasses could determine the method(s) by overwriting the searchKey and
 * searchPattern methods, respectively.
 * By default the two example methods are provided.
 * @author Ari Korhonen
 *
 */

public class SearchTable extends VirtualObject implements CDT, Code {
    private Table text;
    private VirtualInteger time = new VirtualInteger();

    public SearchTable() {
        myKeys e = new myKeys("NOTSOVERYLONGEXAMPLETEXT");
        text = new Table(e.getElements(), 1);
    }

    class myKeys {
        Element[] e;
        public myKeys(String k) {
            e = new Element[k.length()];
            for (int i=0; i<k.length(); i++)
                e[i] = new myKey(k.substring(i,i+1));
        }
        public Element[] getElements() { return e; }

    }

    class myKey extends Key implements matrix.decoration.Styled {
        VirtualBoolean visited = new VirtualBoolean();
        public myKey(String s) { super(s); }
        public matrix.decoration.StyleSheet getStyleSheet() {
            return new matrix.decoration.StyleSheetAdapter() {
                    public java.awt.Color getDefaultBackgroundColor() {
                        if (isVisited())
                            return java.awt.Color.blue;
                        else
                            return java.awt.Color.white;
                    };
                    public java.awt.Color getDefaultPenColor() {
                        if (isVisited())
                            return java.awt.Color.white;
                        else
                            return java.awt.Color.red;
                    };
                };
        }
        public boolean isVisited() { return visited.eval(); }
        public void setVisited(boolean flag) { visited.assign(flag); }
    }

    // (Abstract) Methods and default implementations
    public boolean searchKey(Element e) {
        return linearSearch(text, e);
    }

    public boolean searchPattern(Array a) {
        return BMH(text, a);
    }

    // --- interface Code --- (code for Boyer-Moore-Horspool algorithm only)

    public String getCodeName() {
        return "Algorithm Boyer-Moore-Horspool";
    }

    Statement[] L = null;
    public CodeLine[] getCodeLines() {
        if (L == null) {
            L = new Statement[] {
                new Statement("procedure BMH;"), // 0
                new Statement("begin"), // 1
                new Statement("  for a:=0 to c do d[a]:=m;"), // 2
                new Statement("  for j:=1 to m-1 do d[p[j]]:=m-j;"), // 3
                new Statement("  x:=p[m];"), // 4
                new Statement("  i:=m;"), // 5
                new Statement("  while i<=n do begin"), // 6
                new Statement("    q:=s[i];"), // 7
                new Statement("    if q=x then begin"), // 8
                new Statement("      j:=m-1; "), // 9
                new Statement("      while p[j]=s[i-m+j] and j>0 do j:=j-1;"), // 10
                new Statement("      if j=0 then"), // 11
                new Statement("         write i-m+1 end;"), // 12
                new Statement("    i:=i+d[q] end"), // 13
                new Statement("end") // 14
                    };
            String BoyerMoore = "In the BMH algorithm, Pattern P is scanned across Text S from left to right (the skip loop at lines 7-14), but the actual comparisons between the pattern and the text are performed from right to left (at line 11). The first comparison is between P[m] and S[m]. In the event of a mismatch, if S[m] does not occur at all in P, then the pattern may be shifted safely m places to the right. This is due to the fact that the possibility of P occuring in S starting at any of the first m positions has been ruled out. The next comparison would then be between P[m] and S[2m]. If S[m] does appear in the pattern and its rightmost occurence is P[m-s], then the pattern may be shifted s places to the right, aligning P[m-s] with S[m]. Testing may be resumed by comparing S[m] with P[m+s], i.e. the text index is incremented by s.";
            L[0].setDescription(BoyerMoore);
            L[1].setDescription(BoyerMoore);
            String Increments = "The increments, s, are precomputed and stored in the shift table d. During the search, the table is indexed by the text character causing a mismatch, so a table the size of the alphabet in use is required. In order to be able to visualize the table d, in this example, the size of the alphabet, c, is precomputed and set to the proper size by evaluating the maximum character from P and S.";
            L[2].setDescription(Increments);
            L[3].setDescription(Increments);
            String FirstComparison = "The first comparison is between x = P[m] and S[i = m].";
            L[4].setDescription(FirstComparison);
            L[5].setDescription("The shift table d has been computed and it is not altered in the subsequential main loop. The pattern has aligned with the beginnin of the text. The variable x contains the rightmost character of the pattern. The variable i points to the text position aligned with the rightmost pattern position. Here i=m is satisfied.");
            L[6].setDescription("This is the skip loop.");
            L[7].setDescription("Here i<=n holds. The variable i points to the next position aligned with the rightmost position of the pattern.");
            L[8].setDescription("The variable q contains the text character S[i] that is compared to the rightmost character of the pattern.");
            L[9].setDescription("Here q=x holds, i.e. the rightmost character of the pattern matches with the text. Other pattern positions should be checked starting from P[m-1].");
            L[10].setDescription("At this point 0<j<m holds. If the pattern character P[j] matches with the text, proceed one position to the left.");
            L[11].setDescription("Either P[j]<>S[i-m+j] or j=0 holds.");
            L[12].setDescription("All the pattern characters match with the text, i.e. a match has been found.");
            //L[13].setDescription("At least one of the conditions P[j]<>S[i-m+1], j=0, and q<>x holds. The checking an alignment has been finished. The algorithm is ready to shift to the next alignment. The length of shift is d[q], where q is the rightmost character of the pattern.");
            L[13].setDescription("Define the description and invariants for this code line");
            L[14].setDescription("Finally i>n holds.");
        }
        return L;
    }

    public int getTime() {
        return time.eval();
    }

    public String[] getVariables() {
        String n_text = n.isDefined() ? "n = "+n.eval()+"; " : "n = ?; ";
        String m_text = m.isDefined() ? "m = "+m.eval()+"; " : "m = ?; ";
        String x_text = x.isDefined() ? "x = "+x.eval()+"; " : "x = ?; ";
        String i_text = i.isDefined() ? "i = "+i.eval()+"; " : "i = ?; ";
        String j_text = j.isDefined() ? "j = "+j.eval()+"; " : "j = ?; ";
        String q_text = q.isDefined() ? "q = "+q.eval() : "q = ?.";
        String ints_text = n_text+m_text+x_text+i_text+j_text+q_text;
        String s_text = ((s != null) && (s.array != null)) ? "s:"+s.array.toString() :
            "array s not defined";
        String p_text = ((p != null) && (p.array != null)) ? "p:"+p.array.toString() :
            "array p not defined";
        String d_text = ((d != null) && (d.array != null)) ? "d:"+d.array.toString() :
            "array d not defined";

        return new String[] { ints_text, s_text, p_text, d_text };
    }

    public String[] getVariableDescription() {
        return new String[] { "n is the size of Text S, m is the size of pattern P, x = P[m], index i refers to the S[i] at which point the next comparison occurs, index j is used for shift table initialization (at line 3) and for the actual comparison between the pattern and the text performed from right to left (at line 11), and q = T[i].", "Text S", "Pattern P", "Shift Table D" };
    }

    /* Linear search algorithm implementation. */
    private boolean linearSearch(Table text, Element key) {
        int size = text.size();
        for (int i=0; i<size; i++)
            if (key.equals(text.getObject(i)))
                return true;
        return false;
    }

    private int max(ArrayHandler p) {
        int max = 0;
        for (int i=1; i<=p.length(); i++)
            if (max < p.get(i)) max = p.get(i);
        return max;
    }
    /* Algorithm Boyer-Moore-Horspool
     * Input: s[0..n-1] text, p[0..m-1] pattern
     * Output: Start indeces of all the occurences of P in S.
     */
    private boolean BMH(Array text, Array pattern) {
        //animator.startOperation();
        Animator animator = Animator.getActiveAnimator();
        animator.startOperation();
        CodeLine[] l = getCodeLines();
        time.assign(0);
        for (int ii=0; ii<l.length; ii++) l[ii].stamp(time.eval());
        animator.endOperation();
        l[0].stamp(time.inc());
        /*ArrayHandler*/ s = new ArrayHandler(text);
        /*ArrayHandler*/ p = new ArrayHandler(pattern);
        /*VirtualInteger*/ m.assign(p.length());
        /*VirtualInteger*/ n.assign(s.length());
        int max = (max(s) > max(p)) ? max(s) : max(p);
        s.clearVisited();
        l[1].stamp(time.inc());
        boolean ret = false;
        l[2].stamp(time.inc());
        /*ArrayHandler d = new ArrayHandler(new VirtualArray()); */
        for (int a=0; a <= max; a++) d.set(a, m.eval());
        l[3].stamp(time.inc());
        for (j.assign(1); j.eval()<m.eval(); j.inc())
            d.set(p.get(j.eval()), m.eval()-j.eval());
        l[4].stamp(time.inc());
        /*MyVirtualInteger*/ x.assign(p.get(m.eval()));
        l[5].stamp(time.inc());
        /*VirtualInteger*/ i.assign(m.eval());
        l[6].stamp(time.inc());
        while (i.eval()<=n.eval()) {
            l[7].stamp(time.inc());
            /*VirtualInteger*/ q.assign(s.get(i.eval()));
            //Note.out(this, "i="+i+"; n="+n+"; m="+m+"; x="+x+"; q="+q);
            l[8].stamp(time.inc());
            if (q.eval() == x.eval()) {
                l[9].stamp(time.inc());
                /*VirtualInteger*/ j.assign(m.eval() - 1);
                //Note.out(this, "j="+j);
                l[10].stamp(time.inc());
                while ((j.eval()>0) && (p.get(j.eval()) == s.get(i.eval()-m.eval()+j.eval())))
                    j.dec();
                l[11].stamp(time.inc());
                if (j.eval()==0) {
                    //Note.out(this, "j==0!!!!!!!");
                    l[12].stamp(time.inc());
                    Note.show(this, "Hit at "+(i.eval()-m.eval()+1)); ret = true;
                }
            }
            l[13].stamp(time.inc());
            i.assign(i.eval() + d.get(q.eval(),m.eval()));
            //Note.out(this, "i+="+d.get(q,m));
        }
        l[14].stamp(time.inc());
        animator.endOperation();
        animator.startOperation();
        x.setDefined(false);
        i.setDefined(false);
        j.setDefined(false);
        q.setDefined(false);
        s.clearVisited();
        time.assign(0);
        //animator.endOperation();
        return ret;
    }

    class ArrayHandler {
        Array array;
        ArrayHandler(Array a) { this.array = a; }
        java.util.Vector visited = new java.util.Vector();
        public void set(int index, int value) {
            //Note.out(this, "set "+value+" at "+index);
            array.setObject(String.valueOf(Character.forDigit(value,10)), index);
            //Note.out(this, "For now on get returns " + get(index));
        }
        public Array getArray() { return array; }
        /* return value at specified index */
        public int get(int index) {
            //Note.out(this, "get value at " + index + " for " + array);
            Object value = array.getObject(index);
            if (value == null) {
                Note.show(this, "Array out of bounds: "+index);
                Note.err(this, "Array out of bounds: "+index);
            }
            if (value instanceof myKey) {
                ((myKey)value).setVisited(true);
                visited.addElement(value);
            }
            return Character.getNumericValue(value.toString().charAt(0));
        }
        /* return value at specified index or notFoundValue */
        public int get(int index, int notFoundValue) {
            Object value = array.getObject(index);
            //Note.out(this, "value at "+index+" ="+value);
            if (value == null) return notFoundValue;
            return Character.getNumericValue(value.toString().charAt(0));
        }
        /* return the size of the array */
        public int length() { return array.getLast()-array.getFirst()+1; }
        public void clearVisited() {
            for (int i=0; i<visited.size(); i++)
                ((myKey)visited.elementAt(i)).setVisited(false);
            visited.removeAllElements();
        }
    }

    static int globallineno = 0;
    class Statement implements CodeLine /* , matrix.visual.stylesheet.Styled */ {
        String line = null;
        String description = null;
        VirtualInteger timeStamp = new VirtualInteger();
        int lineno = ++globallineno;
        public Statement(String line) { this.line = line; }
        public String getCodeLine() { return lineno + ". "+ line; }
        public void stamp(int time) {
            Animator animator = Animator.getActiveAnimator();
            timeStamp.assign(time);
            animator.endOperation();
            animator.startOperation();
        }
        public int getStamp() { return timeStamp.eval(); }
        public String getDescription() { return description; }
        public void setDescription(String description) { this.description = description; }
        /*
          public matrix.visual.stylesheet.StyleSheet getStyleSheet() {
          return new matrix.visual.stylesheet.StyleSheetAdapter() {
          //public java.awt.Color getDefaultBackgroundColor() {
          //          if (isVisited())
          //            return java.awt.Color.blue;
          //      else
          //        return java.awt.Color.white;
          //};
          //public java.awt.Color getVisitedBackgroundColor() { // current line color
          //    return java.awt.Color.white;
          //};
          public boolean isInfoEnabled() { return true; }
          public String getInfo() { return "Simon says: "+getDescription(); }
          };
          }
        */
    }

    class StyledVirtualArray extends VirtualArray implements StyledArray {
        public String getIndexName(int i) {
            String ind = String.valueOf(Character.forDigit(i,i+2));
            //Note.out(this, "ind="+ind);
            //return "\"" + ind.toUpperCase() + "\"";
            return ind.toUpperCase();
        }
    }

    ArrayHandler s,p;
    ArrayHandler d = new ArrayHandler(new StyledVirtualArray());
    MyVirtualInteger n = new MyVirtualInteger();
    MyVirtualInteger m = new MyVirtualInteger();
    MyVirtualInteger x = new MyVirtualInteger();
    MyVirtualInteger i = new MyVirtualInteger();
    MyVirtualInteger j = new MyVirtualInteger();
    MyVirtualInteger q = new MyVirtualInteger();
    class MyVirtualInteger extends VirtualInteger implements Primitive, FDT  {
        String name;
        VirtualBoolean isDefined;
        MyVirtualInteger(int i, String name, boolean defined) {
            super(i);
            this.name = name;
            isDefined = new VirtualBoolean(defined);
        }
        MyVirtualInteger() { this(-1, "not defined", false);  }
        public Object getElement() { return primitiveValue; }
        public void setElement(Object o) { /*primitiveValue = o;*/ }
        public String toString() { return name + " == " + eval(); }
        public boolean isDefined() { return isDefined.eval(); }
        public void setDefined(boolean defined) { isDefined.assign(defined); }
        public int assign(int i) { setDefined(true); return super.assign(i); }
    }

    private Element turnElement(Object o) {
        if (o instanceof Element)
            return (Element)o;
        else {
            Note.err(this, "Cannot turn "+o+" to Element");
            return null;
        }
    }

    // --- interface FDT ---

    public Object getElement() {
        return text;
    }

    public void setElement(Object o) {
        if (o instanceof Table)
            text = (Table)o;
        else
            Note.show(this, "Cannot change the text to " + o);
    }

    // --- interface CDT ---

    public FDT insert(Object o) {
        if (o instanceof Element)
            Note.show(this, "Element " + o +
                      (searchKey((Element)o) ? "" : " not") + " found.");
        else if (o instanceof Array)
            Note.show(this, "Pattern " + o +
                      (searchPattern((Array)o) ? "" : " not") + " found.");
        else
            Note.show(this, "Dunno how to search " + o);
        return text;
    }

    public FDT delete(Object o) {
        Note.show(this, "Delete not implemented.");
        return text;
    }

    public Object search(Object o) {
        return this;
    }

    // --- exercise ---
    public Array getD() {
        return d.getArray();
    }

    static final long serialVersionUID = 2307640565853602985L;
}
