/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: CommonGraphImpl.java,v 1.10 2004/05/26 08:05:54 vkaravir Exp $ */

package matrix.structures.FDT.probe;

import java.util.Vector;

import matrix.structures.FDT.*;
import matrix.structures.memory.*;
import matrix.util.Note;

/** A common graph implementation in which every node has implemented
 * as a VirtualArray. The first position (index = DATA = 0) stores the
 * element and the rest of the array stores the successors.
 */
public class CommonGraphImpl

    implements Graph, Vertex {
    // element at position zero
    private final static int DATA = 0;
    private VirtualArray graph;

    /** The constructor for an empty graph. */
    public CommonGraphImpl() {
        graph = new VirtualArray();
    }

    /** The constructor for a new root node. */
    public CommonGraphImpl(Object obj) {
        graph = new VirtualArray();
        setElement(obj);
    }

    // CDT (not actually defined as an interface)

    /** Return the element value. */
    public Object getElement() {
        return graph.getObject(DATA);
    }

    /** Set the element value. */
    public void setElement(Object v) {
        graph.setObject(v, DATA);
    }

    // Methods for interface Array

    public VirtualArray getArray() {
        return graph;
    }

    /** @return object at specified index. */
    public Object getObject(int index) {
        return graph.getObject(index);
    }

    /** Set object at specified index. */
    public void setObject(Object o, int index) {
        if (index == 0)
            setElement(o);
        else
            Note.show(this, "Illegal operation");
    }

    /** @return the value of the first index. */
    public int getFirst() {
        return 0;
    }

    /** @return the value of the last index. */
    public int getLast() {
        return graph.size()-1;
    }

    // Methods for interface Graph

    public int getSubGraphCount() {
        if (getElement() == null) return 0;
        //if ((graph == null) || (graph.size() == 0)) return 0;
        return graph.size() - 1;
    }

    public Graph[] getSubGraphs() {
        Graph[] t = new Graph[getSubGraphCount()];
        for (int i=0; i<t.length; i++)
            t[i] = (Graph)graph.getObject(i+1);
        return t;
    }

    // Methods for interface Graph

    public void setSubGraph(Graph t, int child) {
        graph.setObject(t, child+1);
    }

    public Graph getNewNode(Object key) {
        return new CommonGraphImpl(key);
    }

    public String toString() {
        if (getElement() instanceof Key)
            return getElement().toString();
        else {
            String s = "null";
            if (getElement() != null) s = getElement().getClass().toString();
            return super.toString() + "(" + s + ")";
        }
    }

    public void insert(FDT s) {
        setSubGraph(getNewNode(s), getSubGraphCount());
    }

    // Methods for interface Graph

    private boolean visited = false;

    private void setVisited() {
        visited = !visited;
    }

    public synchronized Vertex[] getVertices() {
        boolean visitedTrue = !visited;
        Vector nodes = new Vector();
        nodes.addElement(this);
        int nextToVisit = 0;
        while (nextToVisit < nodes.size()) {
            CommonGraphImpl node =
                (CommonGraphImpl)nodes.elementAt(nextToVisit++);
            if (node.visited == visitedTrue) continue;
            node.setVisited();
            Vertex[] subGraph = node.getSuccessors();
            for (int i=0; i<subGraph.length; i++)
                nodes.addElement(subGraph[i]);
        }
        Vertex[] v = new Vertex[nodes.size()];
        nodes.copyInto(v);
        return v;
    }
    /*
    public void setVertex(Graph dg, int i) {
        setSubGraph(dg, i);
    }
    */

    public void insertVertex(Object o) {
    }

    public void deleteVertex(Vertex v) {}

    public void insertEdge(Vertex from, Vertex to) {}

    public void deleteEdge(Vertex from, Vertex to) {}

    // Methods for interface Vertex

    public Vertex[] getSuccessors() {
        Graph[] subGraph = getSubGraphs();
        Vertex[] succ = new Vertex[subGraph.length];
        for (int i=0; i<subGraph.length; i++)
            succ[i] = (Vertex)subGraph[i];
        return succ;
    }

    public void addSuccessor(Vertex v) {}

    public void removeSuccessor(Vertex v) {}

    // Methods for interface List
    /*
    private LinkedList next = null;

    public LinkedList getNext() { return next; }

    public LinkedList getPrevious() { return null; }
    */

    static final long serialVersionUID = 4427042463178194026L;
}
