/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: SimulationTree2Impl.java,v 1.9 2004/05/26 08:05:55 vkaravir Exp $ */

package matrix.structures.FDT.probe;

import java.util.Vector;

import matrix.structures.FDT.*;
import matrix.structures.memory.*;
import matrix.util.Note;

/** A common tree implementation in which every node has implemented
 * as a VirtualArray. The first position (index = DATA = 0) stores the
 * element and the rest of the array stores the children.
 */
public class SimulationTree2Impl
    implements FDT, SimulationTree2, SimulationTree, Vertex {
    // element at position zero
    private final static int DATA = 0;
    private VirtualArray tree;

    /** The constructor for an empty tree. */
    public SimulationTree2Impl() {
        tree = new VirtualArray();
    }

    /** The constructor for a new root node. */
    public SimulationTree2Impl(Object obj) {
        tree = new VirtualArray();
        setElement(obj);
    }

    // CDT (not actually defined as an interface)

    /** Return the element value. */
    public Object getElement() {
        return tree.getObject(DATA);
    }

    /** Set the element value. */
    public void setElement(Object v) {
        tree.setObject(v, DATA);
    }

    // Methods for interface Array

    public VirtualArray getArray() {
        return tree;
    }

    /** @return object at specified index. */
    public Object getObject(int index) {
        return tree.getObject(index);
    }

    /** Set object at specified index. */
    public void setObject(Object o, int index) {
        if (index == 0)
            setElement(o);
        else
            Note.show(this, "Illegal operation");
    }

    /** @return the value of the first index. */
    public int getFirst() {
        return 0;
    }

    /** @return the value of the last index. */
    public int getLast() {
        return tree.size()-1;
    }

    // Methods for interface Tree

    public int getSubTreeCount() {
        if (getElement() == null) return 0;
        return tree.size() - 1;
    }

    public Tree[] getSubTrees() {
        Tree[] t = new Tree[getSubTreeCount()];
        for (int i=0; i<t.length; i++)
            t[i] = (Tree)tree.getObject(i+1);
        return t;
    }

    // Methods for interface Dynamic Tree

    public void setSubTree(Tree t, int child) {
        tree.setObject(t, child+1);
    }
    public Tree getNewNode(Object key) {
        return new SimulationTree2Impl(key);
    }

    public String toString() {
        if (getElement() instanceof Key)
            return getElement().toString();
        else {
            String s = "null";
            if (getElement() != null) s = getElement().getClass().toString();
            return super.toString() + "(" + s + ")";
        }
    }

    public void insert(FDT s) {
        setSubTree(getNewNode(s), getSubTreeCount());
    }

    // Methods for interface Graph
    private boolean visited = false;

    private void setVisited() {
        visited = !visited;
    }

    public synchronized Vertex[] getVertices() {
        boolean visitedTrue = !visited;
        Vector nodes = new Vector();
        nodes.addElement(this);
        int nextToVisit = 0;
        while (nextToVisit < nodes.size()) {
            Tree node = (Tree)nodes.elementAt(nextToVisit++);
            if (((SimulationTree2Impl)node).visited == visitedTrue) continue;
            ((SimulationTree2Impl)node).setVisited();
            Tree[] subTree = node.getSubTrees();
            for (int i=0; i<subTree.length; i++)
                nodes.addElement(subTree[i]);
        }
        Vertex[] v = new Vertex[nodes.size()];
        nodes.copyInto(v);
        return v;
    }

    // Methods for interface Vertex

    public Vertex[] getSuccessors() {
        Tree[] subTree = getSubTrees();
        Vertex[] succ = new Vertex[subTree.length];
        for (int i=0; i<subTree.length; i++)
            succ[i] = (Vertex)subTree[i];
        return succ;
    }

    public void addSuccessor(Vertex v) {}

    public void removeSuccessor(Vertex v) {}

    // Methods for interface List
    /*
    private LinkedList next = null;

    public LinkedList getNext() { return next; }

    public LinkedList getPrevious() { return null; }
    */

    static final long serialVersionUID = -8083574636340721352L;
}
