/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: Rotations.java,v 1.12 2004/05/25 13:28:22 kstalnac Exp $ */

package matrix.structures.other;

import matrix.structures.FDT.probe.BinaryTree;


/**
 * <p>
 * Library class to rotate binarytries. Both single- and double rotations 
 * are available: </p>
 * <p> 
 * <code> Root = Rotations.withRightChild(A)</code> performs following operation:
 * <pre>
 *       A              C
 *      / \     =>     / \
 *     B   C          A   E
 *        / \        / \
 *       D   E      B   D
 * </pre></p>

 * <p>rotateWithtLeftChild is mirror symmetric to rotateWithRightChild.</p>
 *
 * <p>Double rotations perform two single rotations and are mirror symmetric together.</p>
 *
 * @see BinaryTree
 * @author Petri Tenhunen
 */
public class Rotations {

    /** 
     * Perform single rotation with left child. 
     * 
     * @param node Node to rotate.
     * @throws NullPointerException if node has no left children.
     */
    public static BinaryTree withLeftChild( BinaryTree node )
    {
	BinaryTree tmp = node.getLeft();
	node.setLeft(tmp.getRight());
	tmp.setRight(node);
	return tmp;
    }

    /**
     * Peforms a single rotation of Splay tree with left child and grandchild.
     *
     * @param tree the root of the tree to be rotated
     * @return the new root of the tree
     */
    public static BinaryTree splayWithLeft(BinaryTree tree) {
	BinaryTree child = tree.getLeft();
	BinaryTree grand = child.getLeft();
	tree.setLeft(child.getRight());
	child.setLeft(grand.getRight());
	grand.setRight(child);
	child.setRight(tree);
	return grand;
    }
	

    /**
     * Peforms a single rotation of Splay tree with right child and grandchild.
     *
     * @param tree the root of the tree to be rotated
     * @return the new root of the tree
     */
    public static BinaryTree splayWithRight(BinaryTree tree) {
	BinaryTree child = tree.getRight();
	BinaryTree grand = child.getRight();
	tree.setRight(child.getLeft());
	child.setRight(grand.getLeft());
	grand.setLeft(child);
	child.setLeft(tree);
	return grand;
    }


    /**
     * Perform double rotation with left child without actually swapping the rotation's
     * root node.
     *
     * @param t Node to rotate.
     */
    public static void rotateDoubleWithLeft_(BinaryTree t) {
      Object tmp = t.getElement();

      // Added this line
      rotateWithRight_(t.getLeft());

      t.setElement(t.getLeft().getElement());
      t.getLeft().setElement(tmp);
      BinaryTree tree = t.getRight();
      t.setRight(t.getLeft());
      t.setLeft(tree);
      tree = t.getRight().getLeft();
      t.getRight().setLeft(t.getRight().getRight());
      t.getRight().setRight(t.getLeft());
      t.setLeft(tree);      
    }      

    
  
    /**
     * Perform single rotation with left child without actually swapping the rotation's
     * root node.
     *
     * @param t Node to rotate.
     */
    public static void rotateWithLeft_(BinaryTree t) {
      Object tmp = t.getElement();
      t.setElement(t.getLeft().getElement());
      t.getLeft().setElement(tmp);
      BinaryTree tree = t.getRight();
      t.setRight(t.getLeft());
      t.setLeft(tree);
      tree = t.getRight().getLeft();
      t.getRight().setLeft(t.getRight().getRight());
      t.getRight().setRight(t.getLeft());
      t.setLeft(tree);      
    }      
    
    /**
     * Rotate binary tree node with right child.
     * 
     * @param node Node to rotate.
     * @throws NullPointerException if node has no right children.
     */
    public static BinaryTree withRightChild( BinaryTree node )
    {
	BinaryTree tmp = node.getRight();
	node.setRight(tmp.getLeft());
	tmp.setLeft(node);
	return tmp;
    }

    /**
     * Perform double rotation with right child without actually swapping the rotation's
     * root node.
     *
     * @param t Node to rotate.
     */
    public static void rotateDoubleWithRight_(BinaryTree t) {
      Object tmp = t.getElement();

      // Added this line
      rotateWithLeft_(t.getRight());

      t.setElement(t.getRight().getElement());
      t.getRight().setElement(tmp);     
      BinaryTree tree = t.getRight();
      t.setRight(t.getLeft());
      t.setLeft(tree);
      tree = t.getLeft().getRight();
      t.getLeft().setRight(t.getLeft().getLeft());
      t.getLeft().setLeft(t.getRight());
      t.setRight(tree);      
    }      

    /**
     * Perform single rotation with right child without actually swapping the rotation's
     * root node.
     *
     * @param t Node to rotate.
     */
    public static void rotateWithRight_(BinaryTree t) {
      Object tmp = t.getElement();
      t.setElement(t.getRight().getElement());
      t.getRight().setElement(tmp);     
      BinaryTree tree = t.getRight();
      t.setRight(t.getLeft());
      t.setLeft(tree);
      tree = t.getLeft().getRight();
      t.getLeft().setRight(t.getLeft().getLeft());
      t.getLeft().setLeft(t.getRight());
      t.setRight(tree);      
    }      
    
    /**
     * Double rotate binary tree node: first left child
     * with its right child; then node <code>node</code> with it's new left child.
     * 
     * @param node Node to rotate.
     * @throws NullPointerException if node has no left children or left children has no
     *         right children.
     */
    public static BinaryTree doubleWithLeftChild( BinaryTree node )
    {
	node.setLeft( withRightChild( node.getLeft() ) );
	return withLeftChild( node );
    }
    
    /**
     * Double rotate binary tree node: first right child
     * with its left child; then node <code>node</code>  with new right child.
     * 
     * @param node Node to rotate.
     * @throws NullPointerException if node has no right children or right children has no
     *         left children.
     */
    public static BinaryTree doubleWithRightChild( BinaryTree node )
    {
	node.setRight(withLeftChild( node.getRight() ));
	return withRightChild( node );
    }
}



