/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package matrix.uitools;

import java.awt.*;
import java.awt.event.*;
import java.awt.print.*;
import java.util.*;

import matrix.animation.*;
import matrix.structures.FDT.FDT;
import matrix.util.*;


public class MatrixFrame extends Frame {
  
  protected static final int X_SPACING = 20;
  protected static final int Y_SPACING = 20;
  protected static int lastXOpen = -1;
  protected static int lastYOpen = -1;
  private static Vector frames = new Vector();
    // common default font
  
  private MenuBar menu;
  private Animator animator;
  private ScrollPane sp;
  private StructurePanel sPanel;
  private VisualAnimator visualAnimator;
  
  public MatrixFrame() {
    this(new Animator());
  }


  public MatrixFrame(Animator a, StructurePanel panel) {
    this(new VisualAnimator(a), a, panel);
  }


  public MatrixFrame(VisualAnimator va, Animator a) {
      this(va, a, new StructurePanel(a));
  }
  

  public MatrixFrame(VisualAnimator va, Animator a, StructurePanel panel) {
    super("Matrix");
    // set default font
    synchronized (frames) {
      frames.addElement(this);
    } 
    setBackground(StructurePanel.getBgColor());
    ApplicationAdapter.setApplication((Application)panel);
    animator = a;

    sPanel = panel;
    visualAnimator = va;
    sPanel.setVisualAnimator(visualAnimator);
    sp = new ScrollPane();
    sp.getVAdjustable().setBlockIncrement(200);
    sp.getVAdjustable().setUnitIncrement(20);
    sp.getHAdjustable().setBlockIncrement(200);
    sp.getHAdjustable().setUnitIncrement(20);
    sp.setBackground(StructurePanel.getBgColor());

    // Set the structurepanel and visualanimator to listen
    // to each other, so that when one changes, the other
    // is also updated.
    sPanel.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent evt) {
	  validate();
	}
      });
    visualAnimator.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent evt) {
	  validate();
	}
      });

    // connect listeners

    this.addWindowListener(new WindowAdapter() {
	public void windowClosing(WindowEvent e) { close(); }
      });
    setPosition();
        
    // layout components
    layoutComponents();

    show();
    menu = createMenuBar();
    setMenuBar(menu);
    //not needed
    //validate(); 
   
  }


  public MatrixFrame(Animator a) {
    this(a, new StructurePanel(a));
  }


  public MatrixFrame(StructurePanel p) {
    this(p.getAnimator(), p);
  }
  
  
  public MatrixFrame(FDT[] structures) {
    this();
    addStructures(structures);
  }
  
  
  public MatrixFrame(FDT[] structures, Animator a) {
    this(a);
    addStructures(structures);
  }


  protected void layoutComponents() {

    // See java.awt.GridBagLayout for details.
    GridBagLayout grid = new  GridBagLayout();
    setLayout(grid); 
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.insets = new Insets(10,10,10,10);
    gbc.fill = GridBagConstraints.NONE;
    add(getVisualAnimator(), gbc);

    gbc.gridy = 1;
    gbc.weightx = 1.0;
    gbc.weighty = 1.0;
    gbc.anchor = GridBagConstraints.NORTH;
    gbc.fill = GridBagConstraints.BOTH;
    ScrollPane sp = getScrollPane();
    sp.add(getStructurePanel());
    add(sp, gbc);
  }

  protected void setPosition() {
    Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
    if (lastXOpen == -1)
      lastXOpen = screen.width/5;
    
    else
      lastXOpen += X_SPACING;
    if (lastYOpen == -1)
      lastYOpen = screen.height/5;
    else
      lastYOpen += Y_SPACING;
    if (lastXOpen + screen.width/2 > screen.width || lastYOpen + screen.height/2 > screen.height) {
      lastXOpen = 0;
      lastYOpen = 0;
    }
    setBounds(lastXOpen, lastYOpen, screen.width/2, screen.height/2);
  }

  
 /**
   * This method must be called whenever this component receives focus.
   */
  public void focusGained() {
    ApplicationAdapter.setApplication(sPanel);
    Animator.setActiveAnimator(getAnimator());
  }


  public void validateComponents() {
    sPanel.validate();
    visualAnimator.validate();
  }


  public void invalidateComponents() {
    sPanel.invalidateComponents();
    visualAnimator.invalidate();
  }

  
  protected void processWindowEvent(WindowEvent e) {
    if(e.getID() == WindowEvent.WINDOW_ACTIVATED)
      focusGained();
    super.processWindowEvent(e);
  }
  
  
  protected MenuBar createMenuBar() {
      	MenuFactory mf = MenuFactory.getInstance();
	return mf.getMenuBar(this);
//      MenuBar mb = new MenuBar();
//      mb.add(new FileMenu(this));
//      mb.add(new FormatMenu(this));
//      mb.add(new InsertMenu(this));
//      mb.add(new OptionMenu(this));
//      mb.add(new GoMenu(this));
//      return mb;
  }
  
  
  public void close() {
    synchronized(frames) {
      frames.removeElement(this);
      if (frames.size() == 0) {
	quit();
      } else {
	dispose();
      }
    }
  }
  

  public void addStructures(FDT[] ss) {
    Application a = getApplication();
    for(int i = 0; i < ss.length; i++) {
      a.addStructure(ss[i]);
    }
  }


  protected Animator getAnimator() {
    return animator;
  }


  protected Application getApplication() {
    return sPanel;
  }

  
  public StructurePanel getStructurePanel() {
    return sPanel;
  }

  
  public ScrollPane getScrollPane() {
    return sp;
  }


  public VisualAnimator getVisualAnimator() {
    return visualAnimator;
  }


  public Frame getNewFrame(StructurePanel p) {
    return new MatrixFrame(p);
  }

  public Frame getNewFrame() {
    return new MatrixFrame();
  }

  public Frame getNewFrame(FDT[] structures, Animator a) {
    return new MatrixFrame(structures, a);
  }

  public void quit() {
    System.exit(0);
  }

  // Methods to handle printing.

  PageFormat pagef = null;
  PrinterJob pj = null;

  public void pageSetup() {
    if (pj == null) 
      pj = PrinterJob.getPrinterJob(); // obtain PrinterJob
    if (pagef == null) 
      pagef = pj.defaultPage(); // obtain default PageFormat
    pagef = pj.pageDialog(pagef); // open the page dialog
  }

  public void print(boolean sequence) {
    if (pj == null) 
      pj = PrinterJob.getPrinterJob(); // obtain PrinterJob
    if (pagef == null) 
      pagef = pj.defaultPage(); // obtain default PageFormat
    if (sequence) // print the animation sequence
      pj.setPageable(new MatrixPageable(this, pagef, sequence));
    else { // print only one frame
      pj.setPrintable(new MatrixPrintable(sPanel), pagef);
    }
    try {
      if (pj.printDialog())
	pj.print();
    } catch(Exception e) {
      //discard error
    }
  }


  public static int getFrameCount() {
    return frames.size();
  }

  
  public static Enumeration frames() {
    return frames.elements();
  }

  public void updateReferences() {
    sPanel.updateReferences();
  }

  public static void main(String[] args) {
    MatrixFrame mf = new MatrixFrame();
    ApplicationAdapter.setApplication((Application)mf.getStructurePanel());
  }
}
