/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: MG_Arc.java,v 1.17 2004/05/26 08:06:02 vkaravir Exp $ */

package matrix.util.export;

import java.awt.*;
import java.io.*;

/**
 * Represents arc in matrix graphics. 
 */
public class MG_Arc extends MG_Element {

    private int x, y, width, height, startAngle, arcAngle;
    private Color fill;

    public MG_Arc(int x, int y, int w, int h, int sAngle, int aAngle, 
		  Color c, Color fill) {
	super();
	this.x = x;
	this.y = y;
	width = w;
	height = h;
	startAngle = sAngle;
	arcAngle = aAngle;
	color = c;
	this.fill = fill;
    }

    public MG_Arc(Rectangle bounds, int sAngle, int aAngle, 
		  Color c, Color fill, int step) {
	this(bounds.x, bounds.y, bounds.width, bounds.height, sAngle,
	     aAngle, c, fill);
	int i = 0;
	setOpacity(0);
	while (i <= step) {
	    addStep(this, i);
	    i++;
	}
    }

    public Rectangle getBounds() {
	return new Rectangle(x, y, width, height);
    }
    public int getStartAngle() { return startAngle; }
    public int getArcAngle() { return arcAngle; }
    public Color getFillColor() { return fill; }

    public MG_Element getStep(int step) {
	MG_Arc temp = (MG_Arc)steps.get(new Integer(step));
	if (temp == null) return null;
	return new MG_Arc(temp.x, temp.y, temp.width, temp.height,
			  temp.startAngle, temp.arcAngle,
			  temp.color, temp.fill);
    }

    public String toString() {
	return "MG_ARC x:" + x + " y:" + y + " width:" + width + " height:" + 
	    height + " startAngle:" + startAngle + " arcAngle:" + arcAngle +
	    " color:" + color + " fill:" + fill;
    }
    
    public String toLatex() {
	float lh = (float)(height/2.0);
	float lw = (float)(width/2.0);
 	float yl = -y - lh;
	float xl = x + lw;
	if (fill != null) {
	    lh--;
	    lw--;
	}
	String tex = "\\move(" + xl + " " + yl + ")\n";
	if (lh == lw) {
	    if (arcAngle >= 0)
		tex += "\\larc r:" + lw + " sd:" + startAngle + " ed:" +
		    (startAngle + arcAngle) + "\n";
	    else  
		tex += "\\larc r:" + lw + " sd:" + (startAngle+arcAngle) +
		    " ed:" + startAngle + "\n";
	} else // elliptical Arcs not yet supported!!
	    tex += "\\larc r:" + lw + " sd:" + startAngle + " ed:" +
		(startAngle + arcAngle) + "\n";
	float gLevel = ColorConverter.toGrayLevel(color);
	if ((gLevel != 0) && (gLevel == 1))
	    tex = "\\setgray 1\n" + tex + "\\setgray 0\n";
	return tex;
    }

    public void toSVG(Writer out, SVGUtil util)
                throws IOException {
	float sh = (float)(height/2.0);
	float sw = (float)(width/2.0);
	//if (width == height) { // circular arc
	    int[] colors = ColorConverter.toRGB(color);
	    String path = createPath(this);
	    out.write("<path d=\"" + path + "\"");
	    if (fill != null) {
		int[] fillc = ColorConverter.toRGB(fill);
		out.write(" fill=\"rgb(" + colors[0] + "," + colors[1] +
			  "," + colors[2] + ")\"" + " stroke=\"none\"");
	    } else
		out.write(" stroke=\"rgb(" + colors[0] + "," + colors[1] + 
			  "," + colors[2] + ")\" fill=\"none\"");
	    if (getOpacity() != 1)
		out.write(" opacity=\"" + getOpacity() + "\"");
	    out.write(">");
	    animate(out, util, path);
	    out.write("</path>\n");
	    //	}
    }

    public int getType() {
	return MG_Element.MG_ARC;
    }

    /**
     * Creates the SVG elements needed to animate this element.
     *
     * @param out Output Writer
     * @param util SVGUtil used exporting this animation
     * @see SVGUtil
     */
    private void animate(Writer out, SVGUtil util, String path) 
	                throws IOException {
	int i = 0;
	MG_Arc curr = this;
	while (hasStep(i+1)) {
	    MG_Arc next = (MG_Arc)steps.get(new Integer(i+1));
	    if (curr.x != next.x || curr.y != next.y || 
		curr.width != next.width || curr.height != next.height ||
		curr.startAngle != next.startAngle ||
		curr.arcAngle != next.arcAngle ||
		(curr.fill != null && next.fill == null) ||
		(curr.fill == null && next.fill != null))
		out.write(util.animate("d", i, createPath(curr), 
				       createPath(next)));
	    if (!curr.color.equals(next.color)) {
		int[] c1 = ColorConverter.toRGB(curr.color);
		int[] c2 = ColorConverter.toRGB(next.color);
		String from = "rgb(" + c1[0] + "," + c1[1] + "," + c1[2] + ")";
		String to = "rgb(" + c2[0] + "," + c2[1] + "," + c2[2] + ")";
		out.write(util.animate("animateColor", "stroke", i, 
				       from, to));
	    }
	    if (curr.fill != null && next.fill != null) {
		if (!curr.fill.equals(next.fill)) {
		    int[] c1 = ColorConverter.toRGB(curr.fill);
		    int[] c2 = ColorConverter.toRGB(next.fill);
		    String from = "rgb(" + c1[0] + "," + c1[1] + "," + 
			c1[2] + ")";
		    String to = "rgb(" + c2[0] + "," + c2[1] + "," +
			c2[2] + ")";
		    out.write(util.animate("animateColor", "fill", i, 
					   from, to));
		}
	    } else if (curr.fill == null && next.fill != null) {
		int[] c2 = ColorConverter.toRGB(next.fill);
		String to = "rgb(" + c2[0] + "," + c2[1] + "," +
		    c2[2] + ")";
		out.write(util.animate("animateColor", "fill", i, 
				       "none", to));
	    } else if (curr.fill != null && next.fill != null) {
		    int[] c1 = ColorConverter.toRGB(curr.fill);
		    String from = "rgb(" + c1[0] + "," + c1[1] + "," + 
			c1[2] + ")";
		    out.write(util.animate("animateColor", "fill", i, 
					   from, "none"));
	    }	
	    if (curr.getOpacity() != next.getOpacity())         // opacity
		out.write(util.animate("opacity", i, "" + curr.getOpacity(), 
				       "" + next.getOpacity()));
	    i++;
	    curr = next;
	}
	float stepLength = util.getStepLength();
	float pause = util.getPause();
	float endTime = (i)*(pause + stepLength) + pause;
	if (!util.addPanel()) {
	    String opac = "0";
	    if (i == util.getMaxStep())
		opac = "0.5";
	    out.write(SVGUtil.animateAdditive("opacity", endTime, stepLength, 
					      opac, "sum"));
	}
	else if (i < util.getMaxStep() && util.addPanel())
	    out.write(util.animate("opacity", i, ""+curr.getOpacity(), "0"));
	
    }

    private static String createPath(MG_Arc arc) {
	float nstartX = (float)(arc.x + arc.width/2.0 + 
		(arc.width/2.0)*Math.cos(2*Math.PI*arc.startAngle/360));
	float nstartY = (float)(arc.y + arc.height/2.0 - 
		(arc.height/2.0)*Math.sin(2*Math.PI*arc.startAngle/360));
	float nendX = (float)(arc.x + arc.width/2.0 +  
	      (arc.width/2.0)*Math.cos(2*Math.PI*(arc.startAngle + 
						   arc.arcAngle)/360));
	float nendY = (float)(arc.y + arc.height/2.0 - 
	      (arc.height/2.0)*Math.sin(2*Math.PI*(arc.startAngle + 
						    arc.arcAngle)/360));
	int largeFlag = (arc.arcAngle > 180) ? 1 : 0;
	int sweepFlag = (arc.arcAngle > 0) ? 0 : 1;
	String fillPath = "";
	if (arc.fill != null)
	    fillPath = " L" + (arc.x + arc.width/2.0) + "," +
		(arc.y + arc.height/2.0) + " z";
	String path = "M" + nstartX + "," + nstartY + " A" +
	    arc.width/2.0 + "," + arc.height/2.0 + " 0 " + 
	    largeFlag + "," + sweepFlag + " " + nendX + ","  + nendY 
	    + fillPath;
	return path;
    }
}
