/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: MG_Rect.java,v 1.18 2004/05/25 13:28:29 kstalnac Exp $ */

package matrix.util.export;

import java.awt.*;
import java.io.*;

/**
 * Represents rectangle in matrix graphics. 
 */
public class MG_Rect extends MG_Element {
    
    private int x, y, width, height, arcWidth, arcHeight;
    private Color fill;

    public MG_Rect(int x, int y, int w, int h, int aWidth, int aHeight,
		   Color color, Color fill) {
	super();
	this.x = x;
	this.y = y;
	width = w;
	height = h;
	arcWidth = aWidth;
	arcHeight = aHeight;
	this.color = color;
	this.fill = fill;
    }

    public MG_Rect(Rectangle rect, int[] arcSize, Color color, Color fill) {
	this(rect.x, rect.y, rect.width, rect.height, arcSize[0],
	     arcSize[1], color, fill);
    }

    public MG_Rect(Rectangle rect, int[] arcSize, Color color,
		   Color fill, int step) {
	this(rect, arcSize, color, fill);
	int i = 0;
	setOpacity(0);
	while (i <= step) {
	    addStep(this, i);
	    i++;
	}
    }
    
    /** Returns the bounds of this element as <code>Rectangle</code>. */
    public Rectangle getBounds() { return new Rectangle(x, y, width, height); }
    /** Returns the x (index 0) and y (index 1) coordinates of this 
     * element as int array.*/
    public int[] getCoord() { return new int[] { x, y }; }
    /** Returns the size of this element. Index 0 is width and 1 is height. */
    public int[] getSize() { return new int[] { width, height }; }
    /** Returns the size of the roundings. Index 0 is arcWidth and 1 is 
     * arcHeight.*/
    public int[] getArcSize() { return new int[] { arcWidth, arcHeight }; }
    /** Returns the fillcolor of this element. */
    public Color getFillColor() { return fill; }

    /** Returns a <code>String</code> representation of this element.*/
    public String toString() {
	return "MG_Rect x:" + x + " y:" + y + " width:" + width + " height:" + 
	    height + " arcWidth:" + arcWidth + " arcHeight:" + arcHeight +
	    " color:" + color + " fill:" + fill;
    }

    // interface MG_Element

    public MG_Element getStep(int step) {
	MG_Rect temp = (MG_Rect)steps.get(new Integer(step));
	if (temp == null) return null;
	return new MG_Rect(temp.getBounds(), temp.getArcSize(),
			temp.getColor(), temp.getFillColor());
    }
    
    public String toLatex() {
	int lh = height;
	int lw = width;
	int yl = -y;
	int xl = x;
	double arcWl = arcWidth/2.0;
	double arcHl = arcHeight/2.0;
	String tex = "";
	float gLevel = ColorConverter.toGrayLevel(color);
	if (arcWidth == 0 || arcHeight == 0) { // sharp edges
	    tex += "\\move (" + xl + " " + yl + ")\n" +
		"\\lvec (" + (xl+lw) + " " + yl + ")\n" +
		"\\lvec (" + (xl+lw) + " " + (yl-lh) + ")\n" +
		"\\lvec (" + xl + " " + (yl-lh) + ")\n" +
		"\\lvec (" + xl + " " + yl + ")\n";
	    if (fill != null) {
		float fillLevel = ColorConverter.toGrayLevel(fill);
		tex += "\\lfill f:" + fillLevel + "\n";
	    } 
	    if (gLevel != 0) 
		tex = "\\setgray " + gLevel + "\n" + tex + "\\setgray 0\n";
	} else { // round edges
	    if (fill == null) {
		double xp = 0;
		String upperLeft = "\\move (" + xl + " " + (yl-arcHl) + ")\n";
		String upperRight = "\\move (" + (xl+width) + " " +
		    (yl-arcHl) + ")\n";
		String lowerRight = "\\move (" + (xl+width) + " " +
		    (yl-height+arcHl) + ")\n";
		String lowerLeft = "\\move (" + xl + " " + 
		    (yl-height+arcHl)  + ")\n";
		for (double i = 0; i <= arcHl; i += 0.2) {
		    double fact = (i*i*1.0)/(arcHl*arcHl*1.0);
		    xp = arcWl*Math.sqrt(1 - fact);
		    upperLeft += "\\lvec (" + (xl + arcWl - xp) + " " + 
			(yl-arcHl+i) + ")\n";
		    upperRight += "\\lvec (" + (xl+width-arcWl+xp) + " " +
			(yl-arcHl+i) + ")\n";
		    lowerRight += "\\lvec (" + (xl+width-arcWl+xp) + " " +
			(yl-height+arcHl-i) + ")\n";
		    lowerLeft += "\\lvec (" + (xl + arcWl- xp) + " " + 
			(yl-height+arcHl-i) + ")\n";
		}
		upperLeft += "\\lvec (" + (xl+arcWl) + " " + yl + ")\n";
		upperRight += "\\lvec (" + (xl+lw-arcWl) + " " + yl + ")\n";
		lowerRight += "\\lvec (" + (xl+lw-arcWl) + " " + 
		    (yl-lh) + ")\n";
		lowerLeft += "\\lvec (" + (xl+arcWl) + " " + 
		    (yl-lh) + ")\n";
		// Draw the sides in order up, right, down, left
		tex += "\\move (" + (xl+arcWl) + " " + yl + ")\n" +
		    "\\lvec (" + (xl+lw-arcWl) + " " + yl + ")\n" +
		    "\\move (" + (xl+lw) + " " + (yl-arcHl) + ")\n" +
		    "\\lvec (" + (xl+lw) + " " + (yl-lh+arcHl) + ")\n" +
		    "\\move (" + (xl+lw-arcWl) + " " + (yl-lh) + ")\n" +
		    "\\lvec (" + (xl+arcWl) + " " + (yl-lh) + ")\n" +
		    "\\move (" + xl + " " + (yl-lh+arcHl) + ")\n" +
		    "\\lvec (" + xl + " " + (yl-arcHl) + ")\n";
		tex += upperLeft + upperRight + lowerRight + lowerLeft;
		if (gLevel != 0)
		    tex = "\\setgray " + gLevel + "\n" + tex + "\\setgray 0\n";
	    } else {
		float fillLevel = ColorConverter.toGrayLevel(fill);
		tex += "\\setgray " + fillLevel + "\n";
		double cx0 = xl + arcWl;
		double cx1 = xl + lw - arcWl;
		double xp = 0;
		for (double i = 0; i <= arcHl; i += 0.2) {
		    double fact = (i*i*1.0)/(arcHl*arcHl*1.0);
		    xp = arcWl*Math.sqrt(1 - fact);
		    tex += "\\move (" + (cx0 - xp) + " " + (yl-arcHl+i) + 
			")\n" + "\\lvec (" + (cx1 + xp) + " " + (yl-arcHl+i) +
			")\n" + "\\move (" + (cx0 - xp) + " " + 
			(yl - lh + arcHl - i) + ")\n" + "\\lvec (" + 
			(cx1 + xp) + " " + (yl - lh + arcHl - i) + ")\n";
		}
		xp = 0;
		tex += "\\move (" + (cx0 - xp) + " " + (yl) + 
		    ")\n" + "\\lvec (" + (cx1 + xp) + " " + (yl) +
		    ")\n" + "\\move (" + (cx0 - xp) + " " + 
		    (yl - lh) + ")\n" + "\\lvec (" + 
		    (cx1 + xp) + " " + (yl - lh) + ")\n";
		tex += "\\move (" + xl + " " + (yl-arcHl) + ")\n" +
		    "\\lvec (" + (xl+lw) + " " + (yl-arcHl) + ")\n" +
		    "\\lvec (" + (xl+lw) + " " + (yl-lh+arcHl) + ")\n" +
		    "\\lvec (" + xl + " " + (yl-lh+arcHl) + ")\n" +
		    "\\lvec (" + xl + " " + (yl-arcHl) + ")\n" +
		    "\\lfill f:" + fillLevel + "\n" +
		    "\\setgray 0\n";
	    }
	}
	return tex;
    }

    public void toSVG(Writer out, SVGUtil util) throws IOException {
	int[] colors = ColorConverter.toRGB(color);
	out.write("<rect");
	out.write(" x=\"" + x + "\"");
	out.write(" y=\"" + y + "\"");
	out.write(" width=\"" + width + "\"");
	out.write(" height=\"" + height + "\"");
	out.write(" stroke=\"rgb(" + colors[0] + "," + colors[1] + "," + 
		  colors[2] + ")\"");
	if (fill != null) {
	    int[] fillc = ColorConverter.toRGB(fill);
	    out.write(" fill=\"rgb(" + fillc[0] + "," + fillc[1] + "," + 
		fillc[2] + ")\"");
	} else {
	    out.write(" fill=\"none\"");
	}
	if (getOpacity() != 1)
	    out.write(" opacity=\"" + getOpacity() + "\"");
	if (arcWidth != 0 || arcHeight != 0) {
	    out.write(" rx=\"" + (arcWidth/2.0) + "\"");
	    out.write(" ry=\"" + (arcHeight/2.0) + "\"");
	}
	out.write(">");
	animate(out, util);
	out.write("</rect>\n");
    }

    public int getType() {
	return MG_Element.MG_RECT;
    }

    // interface MG_Element done

    /**
     * Creates the SVG elements needed to animate this element.
     *
     * @param out Output Writer
     * @param util SVGUtil used exporting this animation
     * @see SVGUtil
     */
    private void animate(Writer out, SVGUtil util) 
                 throws IOException {
	int i = 0;
	MG_Rect curr = this;
	while (hasStep(i+1)) {
	    MG_Rect next = (MG_Rect)steps.get(new Integer(i+1));
	    if (curr.x != next.x) 
		out.write(util.animate("x", i, "" + curr.x, "" + next.x));
	    if (curr.y != next.y)
		out.write(util.animate("y", i, "" + curr.y, "" + next.y));
	    if (curr.width != next.width) 
		out.write(util.animate("width", i, "" + curr.width, 
						    "" + next.width));
	    if (curr.height != next.height)
		out.write(util.animate("height", i, "" + curr.height, 
				       "" + next.height));
	    if (!curr.color.equals(next.color)) {
		int[] c1 = ColorConverter.toRGB(curr.color);
		int[] c2 = ColorConverter.toRGB(next.color);
		String from = "rgb(" + c1[0] + "," + c1[1] + "," + c1[2] + ")";
		String to = "rgb(" + c2[0] + "," + c2[1] + "," + c2[2] + ")";
		out.write(util.animate("animateColor", "stroke", i, 
				       from, to));
	    }
	    if (curr.fill != null && next.fill != null) 
		if (!curr.fill.equals(next.fill)) {
		    int[] c1 = ColorConverter.toRGB(curr.fill);
		    int[] c2 = ColorConverter.toRGB(next.fill);
		    String from = "rgb(" + c1[0] + "," + c1[1] + "," + 
			c1[2] + ")";
		    String to = "rgb(" + c2[0] + "," + c2[1] + "," + 
			c2[2] + ")";
		    out.write(util.animate("animateColor", "fill", i, 
					   from, to));
		}
	    if (curr.getOpacity() != next.getOpacity()) 
		out.write(util.animate("opacity", i, "" + curr.getOpacity(), 
				       "" + next.getOpacity()));
	    if (curr.arcWidth != next.arcWidth) 
		out.write(util.animate("rx", i, "" + (arcWidth/2.0),
				       "" + (next.arcWidth/2.0)));
	    if (curr.arcHeight != next.arcHeight) 
		out.write(util.animate("ry", i,  "" + (arcHeight/2.0),
						 "" + (next.arcHeight/2.0)));
	    i++;
	    curr = next;
	}
	float pause = util.getPause();
	float stepLength = util.getStepLength();
	float endTime = (i)*(pause + stepLength) + pause;
	if (!util.addPanel()) {
	    String opac = "0";
	    if (i == util.getMaxStep())
		opac = "0.5";
	    out.write(SVGUtil.animateAdditive("opacity", endTime, stepLength, 
					      opac, "sum"));
	}
	else if (i < util.getMaxStep() && util.addPanel())
	    out.write(util.animate("opacity", i, ""+curr.getOpacity(), "0"));
    }
}
