/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: MatrixGraphicsContents.java,v 1.20 2004/05/26 08:06:03 vkaravir Exp $ */

package matrix.util.export;

import java.awt.*;
import java.util.*;

/**
 * Represents the contents of a MatrixGraphics -class.
 */
public class MatrixGraphicsContents {

    /** Root of this "tree". */
    private MG_Group root;
    /** Current group where new elements are added. */
    private MG_Group curGroup;
    /** Current level. */
    private int level = 0;
    /** Current scale. */
    private double scale = 1; 
    /** Current font-size. */
    // XXX Can't use different font-sizes in on drawing. Currently Matrix
    // doesn't need it, but perhaps someday??
    private int fontsize = -1;
    /** Minimum value of the x-coordinates used. */
    private int xmin;
    /** Minimum value of the y-coordinates used. */
    private int ymin;
    /** Maximum value of the x-coordinates used. */
    private int xmax;
    /** Maximum value of the y-coordinates used. */
    private int ymax;
    /** Id of the current group. Index 0 is VisualType Id and index 1 
     * is FDT Id
     */
    private String[] id = new String[] { "", ""};
    /** All fdt id's used are stored here in order to make search faster */
    private Vector fdtList = new Vector();
    /** Elements are stored here until their group's id is known */
    private Vector buffer = new Vector();
    /** Current step in animation */
    private int step = 0;
    /** Flag to tell whether the id is set or not */
    private boolean idOk;

    /**
     * Constructs a new instance of <code>MatrixGraphicsContents</code>.
     */
    public MatrixGraphicsContents() {
	xmin = 10000;
	xmax = -10000;
	ymin = 10000;
	ymax = -10000;
	root = new MG_Group();
	String[] rootId = new String[] { "", "" };
	root.setId(rootId);
	curGroup = root;
	idOk = true;
    }
    
    /**
     * Adds a new line object to the current group.
     *
     * @param x1 x coordinate of the starting point
     * @param y1 y coordinate of the starting point
     * @param x2 x coordinate of the end point
     * @param y2 y coordinate of the end point
     * @param color color of the line
     */
    public void addLine(int x1, int y1, int x2, int y2, Color color) {
	add(new MG_Line(x1, y1, x2, y2, color));
	xmax = max(x1, x2, xmax, true);
	xmin = max(x1, x2, xmin, false);
	ymax = max(y1, y2, ymax, true);
	ymin = max(y1, y2, ymin, false);
    }

    /**
     * Adds a new arc object to the current group.
     * 
     * @param x x coordinate of the circles upper left corner
     * @param y y coordinate of the circles upper left corner
     * @param width width of the circle
     * @param height height of the circle
     * @param startAngle starting angle of the arc
     * @param endAngle end angle of the arc
     * @param c color of the arc
     * @param fill fill color of the arc
     */
    public void addArc(int x, int y, int width, int height, int startAngle,
		       int arcAngle, Color c, Color fill) {
	add(new MG_Arc(x, y, width, height, startAngle, arcAngle,
				c, fill));
	xmax = max(x, x + width, xmax, true);
	xmin = max(x, x + width, xmin, false);
	ymax = max(y, y + height, ymax, true);
	ymin = max(y, y + height, ymin, false);
    }

    /**
     * Adds a new oval object to the current group.
     *
     * @param x the x coordinate of the upper left corner of the oval
     * @param y the y coordinate of the upper left corner of the oval
     * @param width the width of the oval
     * @param height the height of the oval
     * @param color the color of the oval
     * @param fill the color the oval is filled with
     */
    public void addOval(int x, int y, int width, int height, Color color,
			Color fill) {
	add(new MG_Oval(x, y, width, height, color, fill));
	xmax = max(x, x + width, xmax, true);
	xmin = Math.min(x, xmin);
	ymax = max(y, y + height, ymax, true);
	ymin = Math.min(y, ymin);
    }

    /**
     * Adds a new text object to the current group.
     *
     * @param str the string of the text object
     * @param x the x coordinate
     * @param y the y coordinate
     * @param color the color of the text
     * @param font the font of the text
     * @param fm the fontmetrics
     */
    public void addText(String str, int x, int y, Color color, Font font,
			FontMetrics fm) {
	xmax = Math.max(x + fm.stringWidth(str), xmax);
	xmin = Math.min(x, xmin);
	ymax = Math.max(y, ymax);
	ymin = Math.min(y - fm.getHeight(), ymin);
	fontsize = font.getSize();
	add(new MG_Text(str, x, y, color, font, fm));
    }

    /**
     * Adds a new rectangle object to the current group.
     *
     * @param x the x coordinate of the upper left corner
     * @param y the y coordinate of the upper left corner
     * @param width the width of the rectangle
     * @param height the height of the rectangle
     * @param arcWidth the horizontal diameter of the arc at the four corners
     * @param arcHeight the vertical diameter of the arc at the four corners
     * @param color the color of the rectangle
     * @param fill the color the rectangle is filled with
     */
    public void addRect(int x, int y, int width, int height, int arcWidth,
			int arcHeight, Color color, Color fill) {
	if (fill != null) {
	    width--;
	    height--;
	}
	add(new MG_Rect(x, y, width, height, arcWidth, arcHeight,
				 color, fill));
	xmax = max(x, x + width, xmax, true);
	xmin = max(x, x + width, xmin, false);
	ymax = max(y, y + height, ymax, true);
	ymin = max(y, y + height, ymin, false);
    }

    /**
     * Adds a new polygon object to the current group.
     *
     * @param xPoints an array of x coordinates
     * @param yPoints an array of y coordinates
     * @param nPoints a total number of points
     * @param color the color of the rectangle
     * @param fill the color the rectangle is filled with
     */
    public void addPolygon(int[] xPoints, int[] yPoints, int nPoints,
			   Color color, Color fill) {
	add(new MG_Polygon(xPoints, yPoints, color, fill));
	for (int i = 0; i < nPoints; i++) {
	    xmax = Math.max(xPoints[i], xmax);
	    xmin = Math.min(xPoints[i], xmin);
	    ymax = Math.max(yPoints[i], ymax);
	    ymin = Math.min(yPoints[i], ymin);
	}
    }

    /**
     * Adds the given element to the current group.
     * 
     * @param elem The <code>MG_Element</code> to be added.
     */
    private void add(MG_Element elem) {
	if (idOk)
	    curGroup.add(elem, step);
	else
	    buffer.addElement(elem);
    }

    /**
     * Moves one level up in the group hierarchy.
     */
    public void levelUp() {
	if (curGroup.equals(root))
	    return;
	curGroup = curGroup.getFather();
	level--;
	String[] curId = curGroup.getId();
	if (curId[0] != null && curId[1] != null && idOk == true) {
	    cutId();
	}
    }

    /**
     * Moves one level down in the group hierarchy. Basically does
     * nothing until you set the Id again for the new group.
     */
    public void levelDown() {
	level++;
	idOk = false;
    }

    /** 
     * Returns the root of this contents group hierarchy. 
     *
     * @return The root of the group hierarchy.
     */
    public MG_Group getRoot() {
	return root;
    }

    /**
     * Returns the bounds of this graphical context.
     *
     * @return The bounds as <code>Rectangle</code>.
     */
    public Rectangle getBounds() {
	return new Rectangle(xmin, ymin, xmax - xmin + 1, ymax - ymin + 1);
    }

    /**
     * Sets the scale of the graphics. Doesn't affect directly
     * the shapes drawn. Currently used only in LaTeX export.
     */
    public void setScale(double scale) {
	this.scale = scale;
    }

    /**
     * Returns the scale of the graphics.
     *
     * @return The scale of the graphics.
     */
    public double getScale() {
	return scale;
    }

    /**
     * Sets current group's id to the given value.
     *
     * @param id The new Id. Index 0 is VisualType Id, index 1 is FDT Id
     * and index 2 is Original FDT's Id.
     */
    public void setId(int[] id) {
	//setting id's correctly
	this.id[0] += ";" + id[0];
	this.id[1] = "" + id[1];
	idOk = true;
	//inserting elements etc.
	MG_Group existing = root.getGroupByVtId(this.id[0]);
	if (existing != null) {
	    curGroup = existing;
	} else { //doesn't exist
	    String temp = this.id[0];
	    temp = temp.substring(0, temp.lastIndexOf(";"));
	    existing = root.getGroupByVtId(temp);
	    curGroup = existing.addGroup();
	    curGroup.setId(this.id);
	    if (step != 0) {
		MG_Group orig = null;
		if (id[1] != -1) {
		    orig = findGroupByFdtId(this.id[1], this.id[0], step - 1);
		}
		if ((orig == null) && (id[2] != -1))
		    orig = findGroupByFdtId("" + id[2], this.id[0], step - 1);
		if (orig != null) {     // original position exists
		    Enumeration en = orig.getContents();
		    while (en.hasMoreElements()) {
			MG_Element origElem = (MG_Element)en.nextElement();
			origElem = origElem.getStep(step - 1);
			if (origElem == null) break;
			origElem.setOpacity(0);
			curGroup.add(origElem, step - 1);
		    }
		}
	    }
	}
	writeBuffer();
	if (!fdtList.contains(this.id[1]))
	    fdtList.addElement(this.id[1]);
    } 

    /**
     * Moves to the next step in animation. Use this method to inform
     * <code>MatrixGraphicsContents</code> that another step in the animation
     * is beginning.
     */
    public void nextStep() {
	step++;
	curGroup = root;
	id = new String[] { "", "" };
	idOk = false;
    }

    /**
     * Sets the name of the object that is currently drawn. 
     *
     * @param name Name of the object.
     */
    public void setName(String name) {
	curGroup.setName(name);
    }

    /**
     * Returns the max or min value of the given values. If max
     * is true maximum value is returned, false otherwise.
     */
    private int max(int num1, int num2, int num3, boolean max) {
	if (max) 
	    return Math.max(Math.max(num1, num2), num3);
	else
	    return Math.min(Math.min(num1, num2), num3);
    }

    /** Cuts the current ID. */
    private void cutId() {
	String temp = this.id[0];
	if (curGroup.getId()[0] == null && curGroup.getId()[1] == null) {
	    return;
	}
	this.id[0] = temp.substring(0, temp.lastIndexOf(";"));
	//curGroup should already be updated
	this.id[1] = curGroup.getId()[1];
    }

    /**
     * Returns the group with the given fdtId. The group's VisualType ID
     * must be different than the given vtId and the group must have
     * the given step.
     *
     * @param fdtId The FDT Id that is looked for.
     * @param vtId The VisualType Id that the group can't have.
     * @param step The step the group must have.
     * 
     * @return The group matching the given properties. Null if no 
     * suitable group is found.
     */
    private MG_Group findGroupByFdtId(String fdtId, String vtId, int step) {
	if (fdtId.trim().equals("") || fdtId == null || step < 0)
	    return null;
	if (!fdtList.contains(fdtId))
	    return null;
	else
	    return root.getGroupByFdtId(fdtId, vtId, step);
    }

    /**
     * Writes the buffer to the current group.
     */
    private void writeBuffer() {
	Enumeration en = buffer.elements();
	while (en.hasMoreElements())
	    add((MG_Element)en.nextElement());
	buffer = new Vector();  //empty buffer
    }

    public int getFontsize() {
	return fontsize;
    }
}
