/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualGraph.java,v 1.45 2004/06/10 12:15:11 kstalnac Exp $ */

package matrix.visual;

import java.awt.*;
import java.util.Stack;

import matrix.structures.CDT.CDT;
import matrix.structures.FDT.*;
import matrix.structures.FDT.probe.*;
import matrix.structures.algorithms.*;
import matrix.util.Note;
import matrix.decoration.LabelDecorator;

/**
 * This is an abstract container class for graphs. The validator is provided
 * that mirrors the actual data structure and validates all the components and
 * references within this container. The protected abstract factory methods 
 * provide a way to construct the concrete components (items and references) 
 * that are used to properly lay out the concrete subclass.
 *
 * @author Ari Korhonen
 * @version $Revision: 1.45 $
 */
public abstract class VisualGraph extends VisualContainer {
    private boolean directed = true;
    /* private boolean symmetricArcs = true; */
    private Vertex[] vertices;

    // used in adding edges to graph
    private boolean edgeAdd = false;
    private VisualGraphComponent edgeFrom;

    //private MenuItem  depth_first;
  //private Menu traverse;

    // private common constructor
    private VisualGraph(FDT fdt) {
        super(fdt);
	width = 40;
	height = 40;
	if(fdt instanceof UndirectedGraph)
	    directed = false;
	MenuItem mi;
    }

    public String getRepresentationName() {
        return "Graph";
    }

    /** Create a new VisualGraph starting from the specified vertex.
     * The representation consists of all the nodes that are reached starting
     * from the specified vertex. 
     */
    public VisualGraph(Vertex v) {
        this((FDT)v);
        vertices = new Vertex[] { v };
    }

    /** Create a new VisualGraph starting from the specified vertices.
     * The representation consists of all the nodes that are reached starting
     * from the specified vertices. 
     */
    /*
    public VisualGraph(Vertex[] vtable) {
        this((FDT)vtable[0]);
        this.vertices = vtable;
    }
    */

    public VisualGraph(Graph g) {
        this((FDT)g);
        vertices = g.getVertices();
    }

    // --- end of costructors ---

    // popUpCheckboxMenuItems

    /** Toggle directed. 
     * @see #isDirected()
     * @see #setDirected(boolean)
     */
    public void Directed() {
        directed = !directed;
        setInvalid();
    }

    /** Return the boolean value of the flag directed. 
     * If the flag is true, this graph should be represented as
     * a directed graph.
     * @see #setDirected(boolean)
     */
    public boolean isDirected() {
        return directed;
    }

    /** Set the boolean value of the flag directed. 
     * @see #isDirected()
     */
    public void setDirected(boolean val) {
        directed = val;
    }

    /** Set the specified vertex's VisualGraphComponent
     ** at the specified position. 
     * If the VisualGraphComponent doesn't exist, create one.
     */
    private void setItem(Vertex v, int position) {
        //Note.out(this, "setItem " +v+" at " +position);
        int p = getItem(v,position);
        if (p == position) return;
        moveItem(p,position);
    }
    
    /** Return true if the specified Vertex has not visited yet. */
    private boolean notVisited(Vertex v, int position) {
        int k = lookUpVisualItem(v, 0);
        return ((k < 0) || (k >= position));
    }

    /** Return the VisualGraphComponent at the specified index. */
    private VisualGraphComponent getVisualGraphComponent(int index) {
        return (VisualGraphComponent)getItem(index);
    }

    /** Set the reference between components at the specified item positions
     * to appear at the specified link position.
     * If no representation is found a new reference is created.
     * In addition, this method adds the components to be each other
     * successor and predecessor.
     */
    private void setLink(int from, int to, int position) {
        //Note.out(this, "set link " + from + "->" + to);
        VisualGraphComponent v1, v2;
        v1 = getVisualGraphComponent(from);
        v2 = getVisualGraphComponent(to);
        int p = getLink(v1,v2,position);
        VisualReference vr = getLink(p);
        v1.addSuccessor(vr);
        v2.addPredecessor(vr);
        //Note.out(this, "p = " + p + "; position = " + position);
        if (p == position) return;
        //Note.out(this, "move " +p+"->"+position);
        moveLink(p,position);
    }
    
    /** @return true if there exists a link between nodes from and to; 
     * false otherwice.
     */
    private boolean linkExists(int from, int to) {
        VisualGraphComponent v1, v2;
        v1 = getVisualGraphComponent(from);
        v2 = getVisualGraphComponent(to);
        int p = lookUpVisualReference(v1,v2,0);
        return (p >= 0);
    }

    /** Return the index of the representation of the specified vertex. 
     * The representation should appear at index greater or equal than the
     * specified position. If no representation is found a new one is
     * created.
     */
    private int getItem(Vertex v, int position) {
        int k = lookUpVisualItem(v, position);
        //Note.out(this, "getItem "+v +"; position = "+k);
        if (k<0) return createVisualItem(v);
        return k;
    }

    /** Return the index of the representation of the reference between
     ** the specified VisualTypes. 
     * The representation should appear at index greater or equal than the
     * specified position. If no representation is found a new one is
     * created.
     */
    private int getLink(VisualComponent v1, VisualComponent v2, int position) {
        int k = lookUpVisualReference(v1,v2,position);
        if (k<0) return createVisualRef(v1,v2);
        return k;
    }

    /** Create a new VisualGraphComponent for the specified Vertex and
     ** return the index of the representation. 
     */
    private int createVisualItem(Vertex v) {
        //Note.out(this, "create new item "+v);
        add(createNewComponent(v));
        return getItemCount()-1;
    }

    /** Create a new VisualReference for the link between the
     ** specified Vertices and return the index of the representation. 
     */
    private int createVisualRef(VisualComponent v1, VisualComponent v2) {
        //Note.out(this, "create new link "+v1+"->"+v2);
        add(createNewReference(v1,v2));
        return getLinkCount()-1;
    }
    
    /** Subclasses should override this method in order to 
     ** return a new concrete VisualGraphComponent for the specified Vertex.
     * By default this method returns a new VisualGraphComponent.
     * @see VisualGraphComponent
     */
    protected VisualGraphComponent createNewComponent(Vertex v) {
        return new VisualGraphComponent(v);
    }

    /** Subclasses should override this method to return a new concrete
     ** VisualReference between the specified Vertices.
     * By default this method returns a new VisualReference
     * @see VisualReference
     */
    protected VisualReference createNewReference(VisualComponent v1, 
                                                 VisualComponent v2) {
        return new VisualReference(v1,v2);
    }
    
    /** Validate this graph.
     * The current representation is verified in order to ensure that
     * this container have a valid visual representation for each node in the
     * actual data structure. 
     */
    public synchronized void validateComponents() {
        if (isDuplicate()) return;
	if(getStructure() instanceof Graph)
	    vertices = ((Graph)getStructure()).getVertices();
        //Note.out(this, "Validate " + vertices + "(" + vertices.length+ ")");
        if ((vertices != null) && (vertices.length > 0)) { // non empty tree
            Stack stack = new Stack();
            for (int i=0; i<vertices.length; i++)
                stack.push(vertices[i]);
            int i = 0; // item position counter 
            int l = 0; // link position counter
            // check that we have visualComponent for each vertex on the graph
	    // valid VisualGraphComponents (VGC) should appear in DFS order
            while (!stack.empty()) { 
                Vertex next = (Vertex)stack.pop();
                // discard empty and visited nodes
                if ((next != null) && (notVisited(next,i))) { 
                    // set the next to be the i'th node of this graph
                    setItem(next, i);
                    // deal with the successors
                    Vertex[] successor = next.getSuccessors();
                    for (int index=0; index<successor.length; index++)
                        if (successor[index] != null) {
                            int vgc = getItem(successor[index],0);
                            if (vgc > i) { // proper successor (not yet seen)
                                stack.push(successor[index]);
                                setLink(i,vgc,l++); // deal with the link
                            } else { // node already dealed with
                                if (isDirected() || !linkExists(vgc,i))
                                    setLink(i,vgc,l++); // deal with the link
                            }
                        }
                    i++;
                }
            }
            // remove unused components && links
            while (i < getItemCount())
                removeItem(i);
            while (l < getLinkCount())
                removeLink(l);
            int count = getItemCount();
            for (int j=0; j<count; j++)
                getVisualGraphComponent(j).endReferenceUpdate();
            
        } else { // "Empty Graph"
            removeAll();
        }
        super.validateComponents();
    }
    
    // --- Dynamic Graph Behaviour ---

    public void write() {       
        getApplication().showMessage("Sorry", "write ["+read+"] invoked: not impemented yet");
    }

    public boolean canInsert() {
	return getStructure() instanceof SimulationGraph || getStructure() instanceof CDT;
    }
    
    public void insert(Object structure) {
	//getApplication().showMessage("Sorry", "insert ["+read+"] invoked: not impemented yet");
	if(getStructure() instanceof SimulationGraph) {
		if (structure instanceof Table) {
			  Table t = (Table)structure;
			  for(int i = 0; i < t.size(); i++)
				((SimulationGraph)getStructure()).addVertex(t.getObject(i));
		} else
	    	((SimulationGraph)getStructure()).addVertex(structure);
	} else if(getStructure() instanceof CDT) {
	    ((CDT)getStructure()).insert(structure);
	}
	else
	    getApplication().showMessage("Sorry", "Cannot insert into non Dynamic Graph" + this);
	setInvalid();
    }

    public void addEdge(VisualGraphComponent vgc) {
	setEdgeAdd();
	setEdgeFrom(vgc);
    }

    public VisualGraphComponent getEdgeFrom() {
	return edgeFrom;
    }

    public void setEdgeFrom(VisualGraphComponent vgc) {
	edgeFrom = vgc;
    }

    boolean getEdgeAdd() {
	return edgeAdd;
    }
    
    void setEdgeAdd() {
	edgeAdd = !edgeAdd;
	setInvalid();
    }

    public void selectEntity(boolean isShiftDown) {
	if(edgeAdd)
	    getApplication().showMessage("Sorry", "Please click on a node");
    }

    public void drag(int x, int y) {
	if(edgeAdd)
	    return;
	else
	    super.drag(x, y);
    }
    

    public void traverse(String cmd) {
	Graph_Traverse trav = TraverseFactory.getTraversingAlgorithm(cmd);
	if(trav == null) {
	    Note.err(this, "No such traversing algorithm found.");
	    return;
	}
	trav.traverse(getStructure());
	debugRefresh();
    }

    public Polygon getMinimizedPolygon(int x, int y, int width, int height) {
        int xpoints[] = new int[4];
        int ypoints[] = new int[4];
        
        xpoints[0] = width/2-1; ypoints[0] = 2;
        xpoints[1] = width/2+1; ypoints[1] = 2;
        xpoints[2] = width; ypoints[2] = height-2;
        xpoints[3] = 1; ypoints[3] = height-2;
        
        return new Polygon(xpoints, ypoints, 4);
    }
    
    public void paint(Graphics g) {
        super.paint(g);
        if (isMinimized()) { paint_minimized(g); return; }
        if (isPicked()) return; // no frames
        paintFrame(g);
    }

    boolean isWeighted = false;
    public boolean weighted() {
	if (getStructure() instanceof DirectedGraphImpl) {
	    isWeighted = ((DirectedGraphImpl)getStructure()).switchWeightEnabled(!isWeighted);

	    Component[] cmp = getLinks();
	    for(int i = 0; i < cmp.length; i++) {
		if (((VisualReference)cmp[i]).isWeighted() != isWeighted) {
		    ((VisualReference)cmp[i]).Weighted();
		}
		((VisualReference)cmp[i]).setInvalid();
	    }

	    return isWeighted;
	}
	return isWeighted;
    }

}
