
/*
 * Matrix -- algorithm simulation system
 * Copyright (C) 2000-2004 Matrix group
 * (See the AUTHORS file and the author tag below for more information
 * on the authors.)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: VisualType.java,v 1.181 2004/06/10 11:03:53 vkaravir Exp $ */

package matrix.visual;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.*;

import matrix.animation.Animator;
import matrix.decoration.*;
import matrix.simulation.*;
import matrix.structures.CDT.CDT;
import matrix.structures.EDT.EDT;
import matrix.structures.FDT.FDT;
import matrix.structures.adaptation.Adapter;
import matrix.structures.memory.Key;
import matrix.uitools.*;
import matrix.util.*;
import matrix.util.export.MatrixGraphics;

/** A VisualType is the abstract superclass for all representations and
 ** their subcomponents. 
 * Several separate functionality could be
 * identified that are briefly described here.
 * 
 * <PRE>- debugging and common functionality -</PRE>
 * Common debugging method dump that could be invoked from the
 * GUI, if provided. Several utility methods, for example, to 
 * determine supercomponents and frames, etc.
 * 
 * <PRE>- GUI operations -</PRE>
 * PopUpMenu handling and GUI dependent functionality.
 * 
 * <PRE>- adding and removing components (items) and links (relations) -</PRE>
 * ---> these methods should be moved to VisualContainer? <---
 * If there is no links in this container the behavior of adding
 * and removing components should be just the same as with java.awt 
 * default methods. However, if there are both components and
 * links, we want to give the look and feel that there is separate 
 * methods for adding and removing those objects from this container. 
 * The default methods operates with components and additional methods 
 * named with "Link" prefix operates with links. In that case, 
 * e.g. getComponentCount() returns the number of items, links
 * are not counted; and getLinkCount() returns the number of
 * links, items are not counted. 
 *
 * <PRE>- repository functionality -</PRE>
 * Repository functionality handles the element addition/deletion
 * to/from container in such a way that we could keep track on which
 * visual objects are visible at the moment and how many representations
 * a structure has active at the moment. This functionality is hidden to
 * standard AWT methods.
 *
 * <PRE>- Animation help features -</PRE>
 * Animation help is handled by the additional (touch) methods
 *
 * <PRE>- ActionListener and ItemListener interfaces -</PRE>
 * <PRE>- Primitive simulation functionality -</PRE>
 * First of all, the VisualType handles all the mouse control events
 * and produces proper method invocations (click, dblClick, etc.) 
 * based on the user actions. On the other hand, VisualType also
 * provides the default functionality or semantics for these methods.
 * For example, click has the "read" semantics and dblClick has the
 * "write" semantics and therefore the corresponding methods are
 * invoked in the default implementation.
 *
 * <PRE>- Drawing and printing funtionality -</PRE>
 * Redesing needed (as with coloring features).
 *
 * @author Ari Korhonen
 * @version $Revision: 1.181 $
 */
public abstract class VisualType extends Container {

    // -- public fields --
    
    /** The FDT this representation is representing. */
    FDT fdt;

    /* stores the clicked VisualType; XXX should be set protected and accessed by getRead() */
    public static VisualType read = new VisualKey(new Key("Unknown"));
    /* Dimension of this VisualType; XXX should be removed and accessed by getBounds()  */
    public int x, y, width, height;
    private Dimension size = new Dimension();
    /* The crane; XXX should be renamed to crane */
    public Picker picker;

    final public static int INSERT_SEMANTICS = 0;
    final public static int SWAP_SEMANTICS = 1;
    /** Current drag&drop semantics that could be one of the following: 
     * INSERT_SEMANTICS, i.e. replace;
     * SWAP_SEMANTICS, i.e. swap with read 
     */
    protected static int semantics = INSERT_SEMANTICS;

    // -- protected fields are visible only for subclasses --

    /* stores the mouseEntered VisualType */
    protected static VisualType entered;
    /* true if this VisualType is picked at the moment */
    protected static boolean picked = false;
    /* true if this VisualType is on focus at the moment */
    protected boolean focus = false;
    /* true if label is shown */
    protected boolean labeled = true;
    /* The popUpMenu attached to this VisualType */
    private PopupMenu popUp;
    /* The Frame in which this VisualType is a subcomponent */
    private Application application;

    protected boolean isVisited = false;

    // Debugger implementation
    private Menu breakpoint;
    
    // -- private fields only visible/usable in this class or by methods --
    private boolean minimized = false;
    private boolean finalized = false;
    private int last_x, last_y;
    // determines whether this visual object responses for simulator
    // events (i.e. drag&drop, click, etc.)
    private boolean simulator = true;
    // determines whether this visual object's subcomponents 
    // response for simulator events (i.e. drag&drop, click, etc.)
    private boolean enabled = true;
    // Tells if addNotify has already been called
    private boolean notifyAlreadyCalled = false;

    /* Tells if an algorithm is running in the fdt this VisualType 
     * is representing, or is a part of
     */
    private Hashtable hashTable; //, minimizedTypes; 
    //contains all the visualizations. FDT as key and VisualType as value
    private static Hashtable visualizations = new Hashtable();
    private Vector removed = new Vector();
    //private static Thread commandThread;
    private int linkCount = 0;
    private int otherComponents = 0;
    // unique id
    private static int cnt = 0;
    //private int nbr = 0;
    public final int nbr;

    private static int scale = 100;

  //protected Animator animator;
  
    private CommandThread fireInfoBox = null;
    private boolean infoAvailable = false; //does the fdt implement InformationDecorator

    // name of chosen representation, this is changed when a new representation
    // is created for the structure
    private String nameOfRepresentation = "";

    /** Debug method. 
     * Subclasses should invoke super.debug() in order
     * to provide all debug information. Every subclass has its own
     * responsibility to provide debug info about its instance variables.
     * The debug info should be printed out with Note.out() utility.
     */
    public void dump(PrintStream out) {
        out.println("fdt="+ ((fdt == null) ? "null" : fdt.getClass().toString()) + 
                 "\n#" + toString() + 
		 " [isDuplicate="+isDuplicate()+
		 "; isValid = "+isValid()+
		 "; minimized = "+minimized+
		 "; alive = "+simulator+
		 "; isVisited = " + isVisited+
		 "; isVisible = " + isVisible()+
		 //"; (x,y) x (width,height) = (" + x + "," + y + ") x (" + 
		 //width + "," + height +")"+
		 "; finalized = "+finalized+ 
		 "; itemCount = "+getItemCount()+ 
		 "; linkCount = "+getLinkCount()+ 
		 "]");
    }

  
    public void dump() {
      //dump(System.out);
      ByteArrayOutputStream buffer = new ByteArrayOutputStream();
      PrintStream out = new PrintStream(buffer);
      dump(out);
      getApplication().showMessage("Dump", buffer.toString());
    }

    
    private Font currFont;


    public static void setScale(int scal) { scale = scal; }
    public static int getScale() { return scale; }

    private static String[] filterLocalMethods(Class clazz) {
        Method[] methods = clazz.getMethods();
        int count = 0;
        for (int i = 0;i<methods.length;i++) {
            if (methods[i].getDeclaringClass().equals(clazz))
                count++;
        }
        String[] ret = new String[count];
        count = 0;
        for (int i = 0;i<methods.length;i++) {
            if (methods[i].getDeclaringClass().equals(clazz)) {
                ret[count++] = methods[i].getName();
	    }
        }
        return ret;
    }

    /** Construct the representation of the specified fundamental data type. */
    public VisualType(FDT t) {
	this.fdt = t;
	size = new Dimension(width = 150, height = 150);
	infoAvailable = (fdt instanceof InformationDecorator);
        // debug
        nbr = cnt++;
        //Note.out(this, "create new representation #" + nbr + " for " + t);
    }

    public String getName() {
        if (super.getName() == null)
            return "Unknown VisualType";
        else
            return super.getName();
    }

    /** Return the name of the structure this visualization is representing.
     * By default, the class name of the FDT is returned.
     * @return the name of the structure this container is representation. 
     */
    public String getStructureName() {
        if (fdt instanceof Adapter)
            return ((Adapter)fdt).getName();
        return parseName(fdt.getClass().toString());
    }

    private String parseName(String name) {
        if (name != null) 
            if (name.substring(0,13).equals("class matrix."))
                return strip(name);
        return name;
    }

    private String strip(String name) {
        if (name != null) {
            int dot = name.lastIndexOf(".");
            if (dot != -1)
                return name.substring(dot+1);
        }
        return name;
    }

    public Animator getAnimator() {
      return getApplication().getAnimator();
    }

    
    /** Return the unique id of this visual representation (debug purposes). */
    public final int getId() { return nbr; }
    public final void setId() { 
	application.showMessage("Internal error", "Cannot change id"); 
    }
    
    /* popUpMenu methods invoked by reflection */
    

    /** This GUI method is invoked by PopUpMenu->reflection().
     * reflect method with no arguments from CDT strored in this visualization 
     */
    public void reflectEDT(String cdtCmd) {
	//Note.out(this,"reflectEDT("+cdtCmd+") begins");
	EDT e = null;
	if (getStructure() instanceof EDT)
	    e = (EDT)getStructure();	    
	else if (getParent() instanceof VisualType) {
	    ((VisualType)getParent()).reflectEDT(cdtCmd);
	    return;
	}
	else if (e == null) {
	    Note.show("", "Illegal operation");
	    Note.out(this, "VisualType not instance of EDT");
	    return;
	}
	Class cl = e.getClass();
	int p1 = cdtCmd.indexOf('(');
	int p2 = cdtCmd.indexOf(')');

	Assert.notFalse(p2 != ++p1,"VisualType->ReflectADT failed"); 
	   
	do {
	    Method[] m = cl.getDeclaredMethods();
	    for (int i=0; i<m.length; i++) {
		if (m[i].getName().equals(cdtCmd)) {
		    Class[] c = m[i].getParameterTypes();
                    if ((c.length == 0)) { 
			// methods with no parameters
			getAnimator().startOperation();
                        try { m[i].invoke(e, null); }
                        catch (Exception ex) { 
			    if(ex instanceof InvocationTargetException) {
				Throwable invoced = ((InvocationTargetException)ex).getTargetException();
				//invoced.printStackTrace());
				//application.showMessage("InvocationTargetException", invoced.toString());
				Note.out(this, "Exception:" + invoced);
				Note.show("", "Illegal operation");
			    } else {
				Note.out(this, "Exception:" + ex.toString());
				Note.show("", "Illegal operation");
				//application.showMessage("Exception", ex.toString()); 
			    }
                        }
                        finally { 
			    getAnimator().endOperation();
			    //Note.out(this,"reflectADT("+cdtCmd+") completed");
			    setInvalid();
			    
			}
			return; 
		    }
		}
	    }
	    cl = cl.getSuperclass();
	} while(cl != null);
	//Note.out(this,"reflectADT("+cdtCmd+") not found in CDT");
    }
    
    /** Get the semantics of behaviour of drag & drop. 
     * @see #setSemantics(int)
     */
    public static int getSemantics() {
	return semantics;
    }

    /** Set the semantics of behaviour of drag & drop.
     *<PRE>
     * if semantics == 
     * - INSERT_SEMANTICS; drop will replace the old value
     * - SWAP_SEMANTICS; drop will swap the old value and read value
     *</PRE>
     */    
    public static final void setSemantics(int sem) {
	if ((sem == INSERT_SEMANTICS) || 
	    (sem == SWAP_SEMANTICS))
	    semantics = sem;
	else
	    Note.err(null, "setSemantics(): unknown value " + sem +
		     "; semantics not changed");
    }

    /** Toggle the semantics between INSERT and SWAP. 
        This method just invokes changeSemantics().
        This method should be disposed and the GUI should be refactored
        if new modes are allowed.
    */
    public static final void toggleSemantics() {
        changeSemantics();
    }
    
    /** This GUI method is invoked by PopUpMenu->reflection().  
     * Change to "next" drag & drop-operation semantics and 
     * notify the user about the change.
     * The loop includes the following semantics:
     * INSERT_SEMANTICS, SWAP_SEMANTICS, -> INSERT_SEMANTICS
     */
    public static void changeSemantics() {
	setSemantics((semantics+1)%2);
    }

    /** This GUI method is invoked by PopUpMenu->reflection().
     * Rename VisualType. 
     */
    public void rename() {
        rename(getApplication().getInput("Rename: ", getName()));
    }

    public void rename(String newName) {
        if (newName.equals(getName()))
            return;
        getAnimator().startOperation();
        setName(newName);
        if (getStructure() instanceof LabelDecorator) {
            LabelDecorator ld = (LabelDecorator)getStructure();
            ld.setLabel(getName());
        }
        getAnimator().endOperation();
        setInvalid();
    }

    /** This GUI method is invoked by PopUpMenu->reflection().
     * Create additional window for this structure. 
     */
    public void newWindow() {
        FDT[] ss = { getStructure() };

	// Panu changed this as a bug fix 2003-01-08
	//Application oldApp =  ApplicationAdapter.getApplication();
	Application oldApp = getApplication();

	//Application newApp = oldApp.getNewInstance(ss,new Animator());
	Application newApp = oldApp.getNewInstance(ss,getApplication().getAnimator());
	oldApp.engage(newApp);
	if (newApp instanceof Frame) {
	  ((Frame)newApp).show();
	} else {
	  newApp.getNewWindow().show();
	}
        newApp.engage((Application)getApplication());
        setInvalid();
    }

    /** This GUI method is invoked by PopUpMenu->reflection().
     * Create additional representation 
     * for this structure within this window. 
     */
    public void newRepresentation() {
        getApplication().addStructure(this.getStructure());
        setInvalid();
    }

    /** This method is invoked (Picker) in order to enable/disable events. 
     * <PRE>
     * b = true invokes enableEvents(mask) 
     * b = false invokes disableEvents(maks)
     * </PRE>
     * @see java.awt.Component#enableEvents
     * @see java.awt.Component#disableEvents
    */
    public void setMask(long mask, boolean b) {
        if (b) enableEvents(mask); else disableEvents(mask);
    }

    // Light Weight Component Handling 

    /** returns recursively the (up most) container 
     * of this components container. 
     */
    public Container getContainer() {
        if (getParent() == null) {
	    Note.err(this, "getContainer: Can't get container"); 
	    return this; 
	}
        if (getParent() instanceof VisualType)
            return ((VisualType)getParent()).getContainer();
        else
            return getParent();
    }

    /** returns the visual container of this component. */
    public VisualContainer getVisualContainer() {
        if (getParent() instanceof VisualContainer) 
            return (VisualContainer)getParent();
        if (getParent() instanceof VisualComponent)
            return ((VisualComponent)getParent()).getVisualContainer();
        return null;
    }


    /** Notifies this component that it has been added to a container. 
     * This method should be called by Container.add, 
     * and not by user code directly.
     * Additional functionality: mouse motion is enabled; the original frame
     * this component is added to is determined; represented structure
     * is stored in the repository and addComponents() is invoked.
     * Subclasses should override addComponents() (and not this!) 
     * in order to add subcomponents for itself. 
     * @see #addComponents()
     * @see #removeNotify()
     */
    public final void addNotify() { 
        super.addNotify(); 
	if(notifyAlreadyCalled)
	  return;
	else
	  notifyAlreadyCalled = true;

        if (getContainer() instanceof Application) {
            this.enableEvents(AWTEvent.MOUSE_EVENT_MASK);
            this.enableEvents(AWTEvent.MOUSE_MOTION_EVENT_MASK);
	    hashTable = new Hashtable();
        } 

        if (getParent() instanceof VisualType) {
            setAlive(enabled = ((VisualType)getParent()).isSubComponentsEnabled());
            hashTable = ((VisualType)getParent()).hashTable;
            application = ((VisualType)getParent()).application;
	    configure(((VisualType)getParent()).getConf());

	    // set the enabled operations for this component
	    VisualTypeConf conf = getConf();
	    if (conf != null) {
		dragEnabled = conf.isEnabled(this.getClass().getName(), VisualTypeConf.DRAG_OPERATION);
		popUpEnabled = conf.isEnabled(this.getClass().getName(), VisualTypeConf.POP_UP_MENU_OPERATION);
		activationEnabled = conf.isEnabled(this.getClass().getName(), VisualTypeConf.HIGHLIGHT_OPERATION);
	    }

        }
        else {
            hashTable = new Hashtable();
	    Container c = getParent();
	    if (c instanceof Application) application = (Application)c;
	    else if (c instanceof Picker) application = (Application)c.getParent();
	    else Note.err(this, "Unknown parent " + c);
        }
        if (this instanceof VisualComponent) {
            FDT ss = getStructure();
            if ((ss != null) && (hashTable != null)) {
                if (!hashTable.containsKey(ss))
                    hashTable.put(ss, this);
            }
        }
        FDT t = getStructure();
        picker = getApplication().getPicker();
	// This is for the visualizations hastable
	Vector vis;
	if (t != null) {    
	    if (visualizations.containsKey(t)) { // t exists in hashtable
		vis = (Vector)visualizations.get(t);
		if (!vis.contains(this))
		    vis.add(this);
	    } else { 
		vis = new Vector();
		vis.add(this);
	    }
	    visualizations.put(t, vis);
	}
        addComponents(); 
        validate(); // XXX should we do this? NO!!!! 
	/* YES! Removing this causes e.g. some table contents to disappear in
	   the prototype UI, and MVT is completely messed up. I recommend
	   further investigation. - JEL */
    }

    protected void setPopUp(PopupMenu pum) {
	popUp = pum;
    }

    protected PopupMenu getPopUp() {
	if (popUp==null) {
	    popUp = MenuFactory.getInstance().getPopupMenu(this);
	    add(popUp);
	}
	return popUp;
    }

    /** Subclasses should override this method in order to add all
     * subcomponents created in the constructor. The constructor fails to 
     * add subcomponents because addNotify() of subcomponent needs to 
     * know the parent of the added component. This information is not 
     * available until the supercomponent is added to it's container. 
     * However, subcomponents created in constructor should usually
     * be added to the container before validate() is invoked first time.
     * addComponents() is invoked after the component is constructed and 
     * before it is first time validated, so subcomponents created 
     * in construction phase could be created here. 
     */
    public void addComponents() {}

    
    /** Notifies this component that it has been removed from a container. 
     * This method should be called by Container.remove, 
     * and not by user code directly.    
     * Additional functionality: remove structure from repository.
     * @see #addNotify()
     */
    public final void removeNotify() {
        FDT ss = getStructure();
        if ((ss != null) && (hashTable != null) && 
	    (hashTable.containsKey(ss))) {
            VisualType t = (VisualType)hashTable.get(ss);
            if (t == this)
                hashTable.remove(ss);
        }
	
	/*
	// XXX Can't remove this visualization from visualizations hastable,
	// since VisualList doesn't work if the visualizations are removed
	if (ss != null) {
	    Vector vis = (Vector)visualizations.get(ss);
	    if (vis != null) {
		System.out.println("removeNotify("+vis.size()+" for "+this);
		if (vis.size() == 1 && vis.contains(this))
		    visualizations.remove(ss);
		else {
		    vis.remove(this);
		    visualizations.put(ss, vis);
		}
	    }
	}*/
        super.removeNotify();
    }

    /** Adds the specified component to this container at the specified index.
     * In addition, the type of the added component is determined and
     * links are treated separately. All items are indexed before links.
     * This method should be called by Container.add, 
     * and not by user code directly.    
     */
    protected /*final*/ void addImpl(Component comp, 
				 Object constraints, 
				 int index) {
        if (removed.contains(comp)) { // adding just removed component
            removed.removeElement(comp);
            comp.setVisible(true);
        }
        if (comp instanceof VisualReference) { // link
            index = 
		index < 0 ? 
		getComponentCount() - otherComponents: 
		getComponentCount() - otherComponents - linkCount + index;
            linkCount++;
        } else if (comp instanceof VisualComponent) { // item
	    index = 
                index < 0 ? 
                getComponentCount() - linkCount - otherComponents: 
                index;
        } else {
            otherComponents++;
            index = getComponentCount();
        }
        super.addImpl(comp, constraints, index);
    }

    /** Change the position of the item between the specified positions. */
    public void moveItem(int from, int to) {
        Component c = getItem(from);
        removeItem(from);
        add(c,to);
    }

    /** Change the position of the link between the specified positions. */
    public void moveLink(int from, int to) {
        Component c = getLink(from);
        removeLink(from);
        add(c,to);
    }

    /** Get the number of items in this container. */
    public int getItemCount() { return getComponentCount() - linkCount - otherComponents; }

    /** Get the number of links in this container. */
    public int getLinkCount() { return linkCount; }

    /** Get the number of other visual components than links and items in this container. */
    public int getOtherCount() { return otherComponents; }

    /** Get the nth item in this container. */
    public VisualComponent getItem(int n) { 
        if (n < getItemCount())
            return (VisualComponent)getComponent(n);
        else 
	    Note.err(this, "getItem():no such item " + n); return null; 
    }

    /** Get the nth link in this container. */
    public VisualReference getLink(int n) {
        if (n < linkCount) {
            int index = getComponentCount() - linkCount - otherComponents + n;
	    return (VisualReference)getComponent(index);
        } else Note.err(this, "getLink():no such link " + n); return null; 
    }

    /** Get the nth non-link-nor-item in this container. */
    public Component getOther(int n) {
        if (n < otherComponents) {
            int index = getComponentCount() - otherComponents + n;
	    return getComponent(index);
        } else Note.err(this, "getOther():no such component " + n); return null; 
    }

    /** Gets all the items in this container. */
    public Component[] getItems() {
        Component[] t = new Component[getComponentCount() - linkCount - otherComponents];
        for (int i=0; i<getComponentCount()-linkCount-otherComponents; i++)
            t[i] = getComponent(i);
        return t;
    }

    /** Gets all the links in this container. */
    public Component[] getLinks() {
        Component[] t = new Component[linkCount];
        for (int i=0; i<linkCount; i++)
            t[i] = getComponent(getComponentCount()-linkCount-otherComponents+i);
        return t;
    }

    /** Removes the component (item or link), 
	specified by index, from this container. */
    public void remove(int index) {
        if ((index >= 0) && (index < getComponentCount())) {
            Component c = getComponent(index);
            if (c instanceof VisualReference) linkCount--;
            else if (!(c instanceof VisualComponent)) otherComponents--;
            removed.addElement(c);
            c.setVisible(false);
            super.remove(index);
        } else Note.err(this, "remove()::Index out of bounds: " + index);
    }

    /** Removes the item, specified by index, from this container. */
    public void removeItem(int index) { 
        if (!(getComponent(index) instanceof VisualComponent))
            Note.err(this, 
		     "removeItem()::no such item " + 
		     index + 
		     ". Object was not a VisualComponent " + 
		     getComponent(index));
        remove(index); 
    }

    /** Removes the link, specified by index, from this container. */
    public void removeLink(int index) {
        if (index < linkCount)
            remove(getComponentCount()-linkCount-otherComponents+index);
        else 
            Note.err(this, "removeLink()::index out of bounds " + index); 
    }

    /** Removes the specified component (item or link) from this container. */
    public void remove(Component c) {
        int count = getComponentCount();
        for (int i=0; i<count; i++)
            if (getComponent(i).equals(c)) {
                remove(i);
                return;
            }
        application.showMessage("Error", "Remove()::Cannot found component " + c);
    }

    /** Removes all the hieararchy components (items and links) from this container. */
    public void removeAll() {
        while ((getItemCount() > 0) || (getLinkCount() > 0))
            remove(0);
    }



    /** Validates this container and all of its subcomponents. 
     * AWT uses validate to cause a container to lay out its subcomponents 
     * again after the components it contains have been added to or modified. 
     * In addition, all the removed components are finalized here by
     * invoking their removeAll method. Thus, subclasses should not
     * override this method but the validateComponents() that is
     * invoked after finalization phase is completed and the container
     * still needs validation.
     */
    public void validate() {
//	Note.out(this, "--- VALIDATE " + this);
        while (!removed.isEmpty()) {
            Container c = (Container)removed.elementAt(0);
            ((VisualType)c).finalized = true;
            c.removeAll();
            c.validate();
            removed.removeElementAt(0);
        }
        if (finalized) { super.validate(); return; }
	boolean doLayoutNeeded = false;
	if (fdt instanceof LabelDecorator) {
            LabelDecorator ld = (LabelDecorator)fdt;
	    doLayoutNeeded = ((getName() == null) || (getName().equals(ld.getLabel())) || (labeled != ld.isLabelEnabled()));
            setName(ld.getLabel());
            labeled = ld.isLabelEnabled();
	} else if (fdt instanceof Styled) {
	    StyleSheet ss = ((Styled)fdt).getStyleSheet();
	    doLayoutNeeded = ((getName() == null) || (getName().equals(ss.getLabel())) || (labeled != ss.isLabelEnabled()));	    
	    setName(ss.getLabel());
            labeled = ss.isLabelEnabled();
	}
	if(doLayoutNeeded) {
	  invalidate();
	  // The return statement caused RedBlackTree and RadixSearchTrie to behave incorrectly.
	  // Nodes and/or labels weren't updated correctly, because validate wasn't 
	  // called. 
	  // The return was orginally added to prevent continous loops. Apparently this
	  // isn't a problem. -jtn
	  //return;
	}
	removePopUpMenu();
        validateComponents();
        for (int i=0; i<getComponentCount(); i++)
            getComponent(i).validate();
        super.validate();
        repaint();
    }

    /** This method is invoked by validate in order to remove the popUpMenu attached into this visual element.
	However, in some elements (i.e. Visual Containers) the popUpMenu should remain, because the
	visual hot spot attached into them employs the same popUp object, thus this method should be
	overridden to do nothing.
    */
    protected void removePopUpMenu() {
	if (popUp != null) {
	    Note.out(this, this + "; popUpMenu removed by validate");
	    remove(popUp); // added in processMouseEvent
	    popUp = null;
	}
    }
    
    /** Validates this container. XXX -> VisualContainer XXX
     * The actual data structure is mirrored and the verification
     * is completed in order to provide one-to-one mapping between
     * the actual data structure and the representation. 
     * The default implementation invoked doLayout()
     */
    public void validateComponents() { }


    /** Causes this container to lay out its components. 
     * Most programs should not call this method directly, 
     * but should invoke the validate method instead. 
     * Containers should override this method in order to lay out 
     * their components.
     */
    public void doLayout() {
//	Note.out(this, "--- DOLAYOUT " + this);
        super.doLayout();
    }

    public void debugRefresh() {
	setInvalid();
	doLayout();
    }

    public PopupMenu createPopUpMenu() {
	//Note.out(this,"Creating new popupmenu");
	PopupMenu p = new PopupMenu();
	p.setEnabled(true);
	return p;
    }

    /** Reflects the simulated method invokation.
     */
    protected void simulatedMethodInvokation(String command) {
        Class cl = getClass();
        do {
            Method[] m = cl.getDeclaredMethods();
            for (int i=0; i<m.length; i++) {
                if (m[i].getName().equals(command)) {
                    Class[] c = m[i].getParameterTypes();
                    // only methods with single parameter allowed
                    if (c.length == 1) { 
                        Object[] parameter = new Object[] { read.getStructure() };
                        try { 
			    getAnimator().startOperation();
			    m[i].invoke(this, parameter); 
			}
                        catch (Exception e) { 
			    if(e instanceof InvocationTargetException) {
				Throwable invoked = 
                                    ((InvocationTargetException)e).
                                    getTargetException();
				invoked.printStackTrace();
				application.showMessage("InvocationTargetException", 
                                                  invoked.toString());
			    } else {
				e.printStackTrace();
				application.showMessage("Exception", e.toString()); 
			    }
                        }
                        finally { getAnimator().endOperation(); setInvalid(); }
			return;
                    }

                }
            }
            cl = cl.getSuperclass();
        } while (cl != null);
        application.showMessage("Error", "Method " + 
                          command + " not found in " + getClass());
    }

    /** Reflects the simulated method invokation.
     */
    protected void simulatedFDTMethodInvokation(String command) {
        Class cl = getStructure().getClass();
        do {
            Method[] m = cl.getDeclaredMethods();
            for (int i=0; i<m.length; i++) {
                if (m[i].getName().equals(command)) {
                    Class[] c = m[i].getParameterTypes();
                    // only methods with no parameters or single parameter allowed
                    if (c.length == 1) { 
                        Object[] parameter = new Object[] { read.getStructure() };
                        try {         
			    getAnimator().startOperation();
			    m[i].invoke(getStructure(), parameter); 
			}
                        catch (Exception e) { 
			    if(e instanceof InvocationTargetException) {
				Throwable invoked = 
                                    ((InvocationTargetException)e).
                                    getTargetException();
				invoked.printStackTrace();
				application.showMessage("InvocationTargetException", 
                                                  invoked.toString());
			    } else {
				e.printStackTrace();
				application.showMessage("Exception", e.toString()); 
			    }
                        }
                        finally { getAnimator().endOperation(); setInvalid();  }
			return;
                    }

		    if (c.length == 0) { 
                        try {         
			    getAnimator().startOperation();
			    m[i].invoke(getStructure(), null); 
			}
                        catch (Exception e) { 
			    if(e instanceof InvocationTargetException) {
				Throwable invoked = 
                                    ((InvocationTargetException)e).
                                    getTargetException();
				invoked.printStackTrace();
				application.showMessage("InvocationTargetException", 
                                                  invoked.toString());
			    } else {
				e.printStackTrace();
				application.showMessage("Exception", e.toString()); 
			    }
                        }
                        finally { getAnimator().endOperation(); setInvalid();  }
			return;
		    }
                }
            }
            cl = cl.getSuperclass();
        } while (cl != null);
        application.showMessage("Error", "Method " + 
                          command + " not found in " + getStructure().getClass());
    }


    /**
       Parses and invokes the reflective method call given as parameter.
    */
    public void reflection(String command) {
        // parse parameter (if exists)
	// Note.out(this, "Calling reflect for command " + command);
        String cmd = command;
        String param = null;
        int p1 = cmd.indexOf('(');
        int p2 = cmd.indexOf(')');
        if (p1 < p2) { 
	    param = cmd.substring(p1+1, p2); 
	    cmd = cmd.substring(0, p1); 
	}
        Class cl = getClass();
        do {
            Method[] m = cl.getDeclaredMethods();
            for (int i=0; i<m.length; i++) {
                if (m[i].getName().equals(cmd)) {
                    Class[] c = m[i].getParameterTypes();
                    if ((c.length == 0) && (param == null)) { 
			// methods with no parameters
                        try { m[i].invoke(this, null); }
                        catch (Exception e) { 
			    if(e instanceof InvocationTargetException) {
				Throwable invoked = ((InvocationTargetException)e).getTargetException();
				invoked.printStackTrace();
				application.showMessage("InvocationTargeException", invoked.toString());
			    } else {
				e.printStackTrace();
				application.showMessage("Exception", e.toString()); 
			    }
                        }
                        finally { setInvalid();  }
			return;
                    } else if ((c.length == 1) && (param != null)) { 
			// methods with single parameter
                        Object[] parameter = new Object[] { param };
                        try { m[i].invoke(this, parameter); }
                        catch (Exception e) { 
			    if(e instanceof InvocationTargetException) {
				Throwable invoked = ((InvocationTargetException)e).getTargetException();
				invoked.printStackTrace();
				application.showMessage("InvocationTargetException", invoked.toString());
			    } else {
				e.printStackTrace();
				application.showMessage("Exception", e.toString()); 
			    }
                        }
                        finally { setInvalid();  }
			return;
                    }
                }
            }
            cl = cl.getSuperclass();
        } while (cl != null);
        application.showMessage("Error", "Method " + command + " not found in " + getClass());
    }

    /** This method invokes setInvalid()
     * @deprecated
     * @see #setInvalid
     */
    public void invalid() {
	setInvalid();
    }

    /** This method invalidates this component and causes all the frames
     * to validate and repaint their representations.
     */
    public void setInvalid() {
        invalidate();
        Application mf = getApplication(); // XXX
        if (mf != null) {
            // notify other frames, too
	    //mf.fireActionEvent(new ActionEvent(this, 0, "invalid"));
            // validate containers in this frame
            mf.validate();
            // validate animator
            mf.validateAnimator();
            // repaint this frame
            mf.repaint();
	  
	    // moved action event here (psilvast & jtn)
	    mf.fireActionEvent(new ActionEvent(this, 0, "invalid"));

        } else Note.warning(this, "setInvalid(): Cannot invalidate");

    }

    /** XXX */
    public Application getApplication() { // Should be MyFrame? 
        if (application != null)
            return application;
	/* Causes a lot of spurious error reports, as the caller has to be
	   prepared for a null value anyway. Removed by JEL.
        Note.err(this, "Error in VisualType " + this + 
		 "; getApplication(): no frame");
        Note.printStackTrace(); //already done in Note.err
	*/
        return null;
    }

    /** Determines the insets of this visual object,
     ** which indicate the size of the object's border. 
     * A VisualContainer, for example, has a top inset that corresponds 
     * to the height of the container's title bar. 
     * @return the default insets of this visual object
     */
    public Insets getInsets() { 
	return new Insets(0,0,0,0);
    }
    
    // simulation and visualisation methods

    private static Object[] touched = new Object[0];
    
    /** This method is invoked in order to visualize 
     ** which elements are involved in some animated action. 
     * Invoked by Assignment of Animator. 
     */
    public static void touch(Object o) {
        if (o != null) {
            Object[] newTouched = new Object[touched.length+1];
            for (int i=0; i<touched.length; i++)
                newTouched[i] = touched[i];
            newTouched[touched.length] = o;
            touched = newTouched;
        } else touched = new Object[0];
    }

    /** This method is invoked in order to visualize 
     ** which elements are involved in some animated action. 
     * Invoked by Assignment of Animator. 
     */
    public static boolean hasTouched(VisualType vt) {
        if (touched != null) {
            for (int i=0; i<touched.length; i++)
                if (touched[i].equals(vt.getStructure()))
                    return true;
        }
        return false; 
    }

    
    /** This method is invoked when visualReference is moved to point
     * to some other visual object. XXX -> VisualReference
     * @param r the reference to be moved
     * @param from the VisualType the reference is moving from
     * @param to the VisualType the reference is moving to
     */
    public void moveReference(VisualReference r, 
			      VisualType from, 
			      VisualType to) {
        application.showMessage("Sorry", "VisualType: Cannot move reference from " + from + " into " + to);
    }

    /** @deprecated replaced by setVisited() */
    public void setMoved(boolean val) {
	setVisited(val);
    }

    /** @deprecated replaced by isVisited() */
    public boolean isMoved() {
        return isVisited();
    }

    public void setVisited(boolean val) {
	isVisited = val;
    }
    /**
     * Determines whether this component has been visited.
     * @return true if this component has been visited; false otherwise.
    public boolean isVisited() {
	return isVisited; 
    }
     */

    /**
     * Determines whether this component has been visited by looking first at the FDT and if it is not
     * Visitable then by returning isVisited flag. 
     * @return true if this component has been visited; false otherwise.
     */
    public boolean isVisited() {
		
	FDT str = getStructure();
	if ((str != null) && (str instanceof Visitable)) {
	    return ((Visitable)str).isVisited();
	}
	return isVisited; 
	
    }

    /** Determines whether this component has the focus. 
     * In addition, component is considered to has the focus 
     * also when it's touched.
     * @see #hasTouched(VisualType)
     */
    public boolean hasFocus() {
	//don't highlight if it isn't enabled
	if (!activationEnabled())
	    return false;
        return (focus || hasTouched(this) || read == this); 
    }

    /** Se focus for this component. Also the original VisualType
     * is set to has the focus in case this component is a duplicate. 
     */
    public synchronized void setFocus(boolean value) {
        focus = value;
        FDT s = getStructure();
        // set focus also for original visualType if this is a duplicate
        if (isDuplicate()) {
            VisualType vt = (VisualType)hashTable.get(s);
            vt.setFocus(value);
        }
        refresh();
    }

    /** Paint this component (and possibly not its subcomponents) 
     ** again without invoking repaint. 
     * Use this method if repaint is too slow possibly because it's
     * too smart logic that combines many repaints into one single
     * update that makes union of the areas to be repainted and thus
     * covers possibly too wide area. Or because repaint paints also
     * all the subcomponents. This method invokes paintFrame that
     * should not repaint subcomponents.
     * @see #paintFrame(Graphics)
     */
    // there is another public void refresh() -method in VisualType,
    // that does something completely different.
    public void refresh() {
        if (getGraphics() != null)
            paintFrame(getGraphics());
        else {
            Note.warning(this, 
                         "Cannot refresh because of no graphics");
	}
    }

    /** Paint (only) the frame of the component.
     * This method is overridden by containers in order to
     * be able to repaint only the border frame and not the
     * components inside the component. By default this
     * method invokes the paint method.
     */
    public void paintFrame(Graphics g) {
        paint(g);
    }
    
    /** Determines whether this component is picked to be drag & dropped. */
    public boolean isPicked() {
        if (picker == null) {
            //Note.err(this, "Missing picker");
            return false;
        }
        return ((picked) && ((read == this) || (this == picker.getPick())));
    }

    /** @return the previously read VisualType. */
    public VisualType getRead() {
        return read;
    }

    /** Determines whether this is the previously read VisualType. */
    public boolean hasRead() {
        return ((read == this) || 
                ((read != null) && 
		 (read.getStructure() == this.getStructure())));
    }

    /** Determines whether this visual object is minimized. 
     * All visual objects are initially not minimized.
     * @return true if the object is minimized; false otherwise. 
     * @see #setMinimized(boolean)
     */
    public boolean isMinimized() {
        return minimized;
    }

    /** Minimizes or maximizes this visual object depending on 
     * the value of parameter b. 
     * @param b - If true, this object should be minimizes; 
     * otherwise, this object should be maximized.
     * @see #isMinimized()
     */
    public void setMinimized(boolean b) {
        minimized = b;
    }
    
    public void minimized() {
	minimized = !minimized;
	setInvalid();
    }

    // enable is an old, deprecated method of java.awt.Component
    public void Enable() {
	enabled = !enabled;
        enableSubComponents(enabled);
        setAlive(true);
	setInvalid();
    }

    public void alive() {
        setAlive(!simulator);
	setInvalid();
    }

    /** Determines whether this visual object is representing an empty object.
     ** An empty object could be "null", "empty subtree", etc. 
     * @return true if the visual object is empty; false otherwise. 
     */    
    public boolean isEmpty() { return false; }

    /** XXX
     * @deprecated isHiding()
     */
    public boolean hidden() { return isHiding(); }

    public boolean isHiding() { return false; }

    /** Determines whether this visual object is representing the
     * specified fundamental data type.
     * Subcomponents should override this method in order to return true
     * whenever this type is representing the specified FDT.
     * The default behaviour is to invoke getSructure() method and determine
     * from that if the actual FDTs are the same.
     * @param fdt - the fundamental data type
     * @return true if this representation is representing the FDT;
     * false otherwise 
     */
    public boolean isRepresenting(FDT s) {
        FDT ss = getStructure();

        //if ((s == null) && (ss == null)) return true;
        //if ((s == null) && (ss != null)) return false;
        //if ((s != null) && (ss == null)) return false;
        return ss == s;
    }
    // debug
    public boolean isRepresenting_(FDT s) {
        boolean r = isRepresenting_(s);
        FDT ss = getStructure();
        return r;
    }

    /** Determines whether this visual object is duplicate representation.
     * @return true if the structure, this representation is representing,
     * already has a drawn representation somewhere else within this 
     * VisualContainer; false otherwise.
     */
    public boolean isDuplicate() {
	FDT s = getStructure();
        if (s == null) return false;
        if (hashTable == null) return false;
        if (hashTable.containsKey(s)) {
            VisualType vt = (VisualType)hashTable.get(s);
            if (this.equals(vt)) 
		return false; 
	    else {
                if (vt.finalized) { // some dead (undone) visualType -> replace
                    hashTable.remove(s);
                    hashTable.put(s, this);
                    return false;
                }
                return true;
            }
        } else hashTable.put(s, this);
        return false;
    }

    /** Determines whether a structure has a representation.
     * @return true if the specified structure
     * already has a drawn representation somewhere else within this 
     * VisualContainer; false otherwise.
     */
    public boolean hasRepresentation(FDT s) {
        if (s == null) return false;
        if (hashTable == null) return false;
        if (hashTable.containsKey(s)) {
            VisualType vt = (VisualType)hashTable.get(s);
            if (this.equals(vt)) 
		return false; 
	    else {
                if (vt.finalized) { // some dead (undone) visualType -> replace
                    return false;
                }
                return true;
            }
        }
        return false;
    }

    /** Gets the preferred size of this component. 
     * @return A dimension object indicating this component's preferred size.
     */
    public Dimension getPreferredSize() {
	//Note.out(this, "--- GETPREFERREDSIZE " + this);
	if ((width != size.width) || (height != size.height))
	    return size = new Dimension(width, height);
	else
	    return size;
    }   

    public void setPreferredSize(int width, int height) {
	size.width = width;
	size.height = height;
    }

    /** Write read object into this object.
     * Subclasses should override this method in order to perform 
     * write operation. Last read visualType component could be located from 
     * instance variable read. Actual storage Structure could be  
     * located by read.getStructure() method invocation. 
     */
    public void write() {
        if (read == null) 
            application.showMessage("Sorry", "No read value");
        else
            application.showMessage("Sorry", "Cannot assign " + 
		      read.getStructure() + 
		      " into " +
		      this.getStructure());
    }

    /** This method is invoked whenever the user clicks this visual object. 
     * The given parameter implies whether the shift key was pressed during the click.
     * Subclasses could override this method if default behaviour 
     * is not valid. By default this method does nothing.
     */
    public void selectEntity(boolean isShiftDown){
    }

     /** This method is invoked whenever the user clicks this visual object.
     * The default behaviour is to call read() method and
     * set this component as read.
     * In addition this method calls the selectEntity method that can be overridden
     * in case default behaviour is not valid.
     */
    private void click() {
 	// if activation is not enabled, call parents click()
	if (!activationEnabled() && getParent() instanceof VisualType) {
	    ((VisualType)getParent()).click();
	    return;
	} else if (!activationEnabled())
	    return;
	if (isEmpty()) return;
        VisualType old = read;
        read = this;        
        if (old != null) old.refresh();         
	selectEntity(isShiftDown);
	setInvalid();
        //refresh();
    }   

    /** This method is invoked whenever the user double clicks 
     ** this visual object. 
     * Subclasses could override this method if default behaviour is not valid.
     * The default behaviour is to call write() method of this component.
     */
    public void dblClick() { 
	// causes possibly problems with new animator design, thus
	return; // this operation is omitted from current release
	//getAnimator().startOperation();
        //entered.invoke(); // write();
	//getAnimator().endOperation();
	//invalid();

    }

    /** This method is invoked after mouse button is pressed.
     * Subclasses could override this method if default behaviour is not valid.
     * The default implementatiion does nothing.
     */      
    public void pressed() {}   

    /** Return true if parameter p is part of this visualtype. 
     * Known use: VisualType.released()
    private boolean partOf(VisualType p) {
    Note.out(this, "this = " + this);
	Note.out(this, "p = " + p);
	if (p.equals(this))
	    return true;
	else if (getParent() instanceof VisualType) {
	    return ((VisualType)getParent()).partOf(p);
	} else return false;
    }
    */

    public boolean isShiftDown = false;


    /** This method is invoked whenever the mouse button 
     * is released after a drag.
     * Subclasses could override this method if default behaviour is not valid.
     * The default behaviour is to invoke insert or swap 
     * methods of this component, depending on the defined semantics.
     * @see #getSemantics()
     */      
    public void released() {
	//Note.out(this, "this = " + this);
	//Note.out(this, "entered = " + entered);
	if (!picked)
	    return;
	// do the drop

        if (isShiftDown) {
            if(!(isDroppedIntoSourceContainer())) {
                if (this.getParent() instanceof VisualType) {
                    VisualType vttemp = this;
		    while(!(vttemp.getParent() instanceof VisualContainer))
			vttemp = (VisualType)vttemp.getParent();
		    vttemp = (VisualType)vttemp.getParent();
                    FDT fdttemp = vttemp.getStructure();
                    if (fdttemp instanceof CDT) {
                        getAnimator().startOperation();
                        ((VisualContainer)vttemp).delete(this.getStructure());
                        getAnimator().endOperation();
                    }
                }
            }
	}

        if (entered != null && dropEnabled()) {
	    doTheDrop();

	}
	// drop to parent
	else if (entered != null && !dropEnabled()) {
	    
	    if (entered.getParent() instanceof VisualType) {
		entered = (VisualType)entered.getParent();
		released();
	    } else {

		application.showMessage("Sorry", 
				  "cannot insert (" + 
				  this + 
				  "): cannot drop on ("+entered+"). ");
	    }
	
	} else {
//  	    application.showMessage("Sorry", 
//  			      "Cannot insert (" + 
//  			      this + 
//  			      "): no target selected ("+entered+"). ");
	}
	picked = false;
	picker.drop();
	setInvalid();
    }


    private boolean isDroppedIntoSourceContainer() {
	VisualType vttemp1 = this;
	VisualType vttemp2 = entered;
	while(vttemp1.getParent() instanceof VisualType) {
	    vttemp1 = (VisualType)vttemp1.getParent();
	}
	if (vttemp2 != null) {
	    while(vttemp2.getParent() instanceof VisualType) {
		vttemp2 = (VisualType)vttemp2.getParent();
	    }
	}
	return (vttemp1 == vttemp2);
    }
    

    private void doTheDrop() {
	picked = false;
	picker.drop();
	
	switch (getSemantics()) {
	case INSERT_SEMANTICS: {
	    //animator.startOperatiOn(new Evaluation(read)); 
	    //Evaluations are removed because in Evaluation there is a reference
	    //to a VisualType which cannot be serialized and therefore saved.
	    getAnimator().startOperation(); 
	    entered.invoke(); 
	    getAnimator().endOperation();
	    break;
	}
	case SWAP_SEMANTICS: {
	    getAnimator().startOperation();
	    entered.swap(); 
	    getAnimator().endOperation();
	    break;
	}
	default: {
	    Note.err(this, "released(): unknown semanics " + 
		     getSemantics());
	    //getAnimator().startOperation(new Evaluation(read)); 
	    getAnimator().startOperation();
	    entered.invoke();
	    getAnimator().endOperation();
	}
	}
	setInvalid();
    } 



    /** This method is invoked when mouse button is entered to this component.
     * Subclasses could override this method if default behaviour is not valid.
     * The default behaviour is to set entered = this and set focus flag on.
     */      
    public void entered() {
	// if activation not enabled -> activate parent
	if (!activationEnabled && getParent() instanceof VisualType) {
	    ((VisualType)getParent()).entered();
	    return;
	}
      //if (fdt instanceof Styled && ((Styled)fdt).getStyleSheet().isInfoEnabled()) {
      if ((fdt instanceof Styled) && (((Styled)fdt).getStyleSheet().isInfoEnabled())) {
	  //Note.out(this, "launching infoBox");
	  int waitTime = ((Styled)fdt).getStyleSheet().getWaitTime();
          fireInfoBox = new CommandThread(waitTime, new Command() { 
                  public void execute() {

		      // Panu changed this as a bug fix 2003-01-08
                      //ApplicationAdapter.
		      //getApplication().
		      //showInfoMessage(VisualType.this,((Styled)fdt).getStyleSheet().getInfo());
		      getApplication().showInfoMessage(VisualType.this,((Styled)fdt).getStyleSheet().getInfo());
                      
                  }
              });
          fireInfoBox.start();
      }
      entered = this; 
      if (!activationEnabled)
	  return;
      setFocus(true); 
    }

    /** This method is invoked when mouse button is exited from this component.
     * Subclasses could override this method if default behaviour is not valid.
     * The default behaviour is to set entered = null and set focus flag off.
     */  
    public final void exited() {
	// if activation not enabled -> activate parent
	if (!activationEnabled && getParent() instanceof VisualType) {
	    ((VisualType)getParent()).exited();
	    return;
	}
        if (fireInfoBox!=null) {
	    //Note.out(this, "removing InfoBox"); 
            fireInfoBox.interrupt();
	    fireInfoBox = null;

	    // calling ApplicationAdapter caused a bug in one exercise. Is it
	    // needed at all? VisualType has reference to application in frame variable.
	    // psilvast: 2003-01-03
	    //ApplicationAdapter.getApplication().removeInfoMessage(this);
	    getApplication().removeInfoMessage(this);
        }
        entered = null; 
        setFocus(false); 
    }
    
    private VisualType drag;
    private static VisualType pickedType;

    /** This method is invoked whenever this visual object is dragged into
     * the specified point.
     */ 
    public void drag(int x, int y) {
        picker.drag(x, y);
    }
    
    private void scroll(ScrollPane sp, int dx, int dy) {
	Point p=sp.getScrollPosition();
	p.translate(dx,dy);
	sp.setScrollPosition(p);
    }

    /** Gets the location of this component in the form 
        of a point specifying the frame's top-left corner.
        The location will be relative to the frame's
        coordinate space.
    */
    public Point getFrameLocation() {
        Point p1 = getContainer().getLocationOnScreen();
        Point p2 = getLocationOnScreen();
        return new Point(p2.x-p1.x, p2.y-p1.y);
    }

    /** This method is invoked whenever this visual object is picked at
     * specified point. 
     */
    public void pick(int x, int y) {
	picked = true;
	read = this;
	drag = this;
	//pickedType = picker.pick(this, getLocationOnScreen(), x, y);  
	pickedType = picker.pick(this, getFrameLocation(), x, y);  
    }
    
    /** This method is invoked when Picker is trying to determine 
     * the relative position in which the cursor should be pointing
     * while the VisualType is dragged over the display. 
     * By default this method returns null indicating that the relative
     * position should be the very same position the user is pointing to 
     * when the drag begins. However, subclass may override this method 
     * to determine other Point.
     * @return A Point, relative to the upper left corner, 
     * that is the correct cursor position. 
     */
    public Point getCursorPoint() {
        return null;
    }

    /**
     * This method is invoked when Picker is trying to determine the
     * set of moving components for drag & drop. By default this
     * method returns null. Subclasses can override this if they want
     * a different representation for objects that are being dragged.
     * @return the VisualType that is dragged on the screen.
     */
    public VisualType getMovingPart() {
	return null;
    }

    /** Enables or disables the simulator operations for this object.
     * By default, all objects response for the simulation operations.
     * However, if this object should not response for the simulation
     * events, the behaviour can be disabled by invoking this method
     * with parameter false. The behaviour can be enabled again by
     * invoking this method with parameter true.
     *  */
    public void setAlive(boolean val) {
        simulator = val;
        if (simulator) {
            this.enableEvents(AWTEvent.MOUSE_EVENT_MASK);
            this.enableEvents(AWTEvent.MOUSE_MOTION_EVENT_MASK);
        } else {
            this.disableEvents(AWTEvent.MOUSE_EVENT_MASK);
            this.disableEvents(AWTEvent.MOUSE_MOTION_EVENT_MASK);
        }
    }

    public boolean isAlive() {
	return simulator;
    }

    /** Return true if subcomponents are enabled for this visual type. */
    public boolean isSubComponentsEnabled() {
        return enabled;
    }

    /** Recursively invokes setSimulator(val) for all 
     * VisualType subcomponents. */
    public void enableSubComponents(boolean val) {
        enabled = val;
        setAlive(val);
        for (int i=0; i<getItemCount(); i++)
            getItem(i).enableSubComponents(val);
        for (int i=0; i<getLinkCount(); i++)
            getLink(i).enableSubComponents(val);
    }


    
    /** Processes mouse events occurring on this component 
     * by dispatching them to the corresponding methods.
     * <PRE>
     * - MOUSE_ENTERED -> entered()
     * - MOUSE_EXITED -> exited()
     * - MOUSE_PRESSED -> pressed()
     * - MOUSE_RELEASED -> released()
     * - MOUSE_CLICKED -> click() or dblClick()
     * </PRE>
     */
    public void processMouseEvent(MouseEvent e) {
        //removes the information about touched objects once 
        // mouse event happens -> elements touched in last
        // redo/undo don't stay in focus after mouse has
        // been on them
        // XXX should this be handled elsewhere?

        if(touched != null)
	  touch(null); 
	try {
	    isShiftDown = e.isShiftDown(); 
	    if (e.isPopupTrigger() && popUpEnabled() && simulator) {
		if (popUp == null) {
		    popUp = MenuFactory.getInstance().getPopupMenu(this);
		    // Note.out(this, "popUpTriggered -> add popUp for "  + this);
		    add(popUp); // removed in validate
		}
		popUp.show(this, e.getX(), e.getY()); 
	    }
	    else if (e.getID() == MouseEvent.MOUSE_ENTERED) {
		entered(); }
	    else if (e.getID() == MouseEvent.MOUSE_EXITED) { 
		exited(); } 
	    else if (e.getID() == MouseEvent.MOUSE_PRESSED) {
		last_x = (short)e.getX(); last_y = (short)e.getY(); pressed(); }
	    else if (e.getID() == MouseEvent.MOUSE_RELEASED) {
		last_x = (short)e.getX(); last_y = (short)e.getY(); released(); }
	    else if (e.getID() == MouseEvent.MOUSE_CLICKED) {
		last_x = (short)e.getX(); last_y = (short)e.getY();
		if ((e.getClickCount() != 1) || (e.isControlDown())) dblClick(); 
		else click();
	    }
	    else {
		super.processMouseEvent(e);
	    }
	} catch (Exception ex) {
	    ex.printStackTrace();
	    application.showMessage("Exception", ex.toString());
	}
    }

    /** Processes mouse motion events occurring on this component 
     * by dispatching them to the corresponding methods.
     * MOUSE_DRAGGED -> drag()
     * MOUSE_MOVED -> mouseMove() !!! DISABLED at the moment !!!
     * MOUSE_MOVED events are not handled because there is no
     * need for them at the moment.
     */
    public void processMouseMotionEvent(MouseEvent e) {
        if (e.getID() == MouseEvent.MOUSE_DRAGGED) {
            if (getToolkit().getSystemEventQueue().
                peekEvent(MouseEvent.MOUSE_DRAGGED) == null) {
                last_x = (short)e.getX(); last_y = (short)e.getY();
                if (!picked && dragEnabled()) pick(last_x, last_y); 
		else if (dragEnabled()) {
		    drag(last_x, last_y);
		    ScrollPane sp;
		    if (getApplication() instanceof StructurePanel) {
			StructurePanel stp=(StructurePanel)getApplication();
			if ((stp.getParent() instanceof ScrollPane)&&
			    (picker!=null)&&
			    (picker.getPick()!=null)) {
			    sp=(ScrollPane)stp.getParent();

			    /* Get scroll pane-relative coordinates... */
			    int x=picker.getPick().getX()+picker.getDX()-
				(int)sp.getScrollPosition().getX();
			    int y=picker.getPick().getY()+picker.getDY()-
				(int)sp.getScrollPosition().getY();
				
			    /* If outside edge, scroll... */

			    if (x<0)
				scroll(sp, -5, 0);
			    if (x>=sp.getViewportSize().getWidth())
				scroll(sp, 5, 0);
			    if (y<0)
				scroll(sp, 0, -5);
			    if (y>=sp.getViewportSize().getHeight())
				scroll(sp, 0, 5);
			}
		    }
		}
            } 
        }
        /*else if (e.getID() == MouseEvent.MOUSE_MOVED) {
          mouseMove(e.getX(), e.getY());
          } */
        else super.processMouseMotionEvent(e);
    }

    // ----- abstract methods ----- 

    /** Determines whether this component contains the specified point. 
     * Subclasses should override this method in order to gain the
     * default border detection in non-rectangular representations.
     * This default implementation does nothing but invokes the
     * super.contains(x,y).
     */
    public boolean contains(int x, int y) {
        return super.contains(x,y);
    }

    /** Returns the FDT this visual object is representing. 
     */
    public FDT getStructure() {
        return fdt;
    }

    /** Set the FDT this visual object is representing. */
    public void setStructure(FDT fdt) {
        this.fdt = fdt;
    }

    /** @return unique FDT (Key) for given Object. 
     * XXX In addition, the future implementation of this method should
     * always return the same FDT for specified Object. XXX
     */
    public FDT turnFDT(Object o) {
        if (o instanceof FDT)
            return (FDT)o;
        else
            return new Key(o);
    }

    // ---------------------------- 
    
    /* Simulation methods */

    private String callee = "assign";

    public Menu createCallMenu() {
	Menu call = new Menu("Call");
	if (getStructure() instanceof 
	    matrix.structures.simulationextensions.CallableMethods) {
	    String[] items = ((matrix.structures.simulationextensions.
		      CallableMethods)getStructure()).
		getCallableMethodNames();	    
	    
	    for (int i=0; i<items.length; i++) {
		MenuItem mi;
		call.add(mi=new MenuItem(items[i]));
		//simulatedMethodInvokation(String command) {
		mi.setActionCommand(items[i]);
		mi.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
			    simulatedFDTMethodInvokation(e.getActionCommand());
			}
		    });
	    }
	}
	return call;
    }

    /** This method is used to determine the default operation to
     * be called whenever method invokation is performed during
     * the simulation. By default, this method returns "assign".
     */
      public String getDefaultCallee() {
          return callee;
      }

      public void setDefaultCallee(String c) {
  	//this sets corresponding checkbox in method menu to true
	  //Note.out(this, "Setting default callee to " + c);
	  //  	for(int i=0; i < method.getItemCount(); i++) {
	  // CheckboxMenuItem cmi = ((CheckboxMenuItem)method.getItem(i));
	  // cmi.setState(c.equals(cmi.getLabel()));
	  //}

          callee = c;
      }

    /** This method is invoked whenever an assign operation is
     * performed during the simulation process. By default, when an
     * object is drag & dropped into this visual object. 
     */
    public abstract void assign(); 
    
    /** This method is invoked when an operation invokation is
     * performed during the simulation process. By default, this
     * method invokes assign() method. Subclasses could override
     * this method in order to provide additional functionality.
     * See, for example, VisualContainer.
     */
    public void invoke() {
	if (getDefaultCallee() == "")
              Note.show(this, "Operation disabled: No default method to invoke");
          else if (getDefaultCallee().equals("assign")) {
              assign();
	  } else 
	    simulatedMethodInvokation(getDefaultCallee());
    }

    /** This method is invoked whenever this visual object is swapped
     * with another object. 
     */
    public void swap() {
	// source should also be in swap mode
        if (getSemantics() != SWAP_SEMANTICS) { 
            invoke();
        } else if (read instanceof VisualKey) {
            VisualKey vk = new VisualKey((Key)read.getStructure());
            VisualType tmp = read;
            if (this instanceof VisualComponent)
		read = getVisualType();
	    else
		read = this;
            tmp.invoke();
            read = vk;
            this.invoke();
        } else application.showMessage("Sorry", "Cannot swap with " + 
                                 read.getStructure());
    }
    
    /** This method is invoked by popUpMenu. */
    private void insert(String key) {
        getAnimator().startOperation();
        read = new VisualKey(new Key(key));
        invoke();
        getAnimator().endOperation();
        setInvalid();
    }
    
    /** This method is invoked whenever the user deletes this visual object. 
     * There is no default implementation, thus the user is notified
     * that this feature is not implemented yet.
     */
    public void erase() {
        application.showMessage("Sorry", "Delete not implemented yet for "+this);
    }

    /** XXX -> VisualComponent? */
    public void remove() {
        application.showMessage("Sorry", "Dunno how to delete key");
    }

    /** XXX -> VisualComponent? */
    public int getReferenceNumber(VisualType to) {
        return 0;
    }

    /** Invoke erase() as a single animation operation. */
    private void erase_() {
        getAnimator().startOperation();
        erase();
        getAnimator().endOperation();
    }

    /** 
     * Removes a key from the structure. If the structure is an CDT
     * this method will call the CDT's delete -method. 
     * XXX --> VisualComponent? */
    public void removeKey(Object o) {
        getAnimator().startOperation();
	removeKeyHelp(o);
	//remove();
	getAnimator().endOperation();
	setInvalid();
    }

    // helper function for removeKey.
    protected void removeKeyHelp(Object o) {
	if(getParent() instanceof VisualType)
	    ((VisualType)getParent()).removeKeyHelp(o);
	else if(getStructure() instanceof CDT) 
	    this.setStructure(((CDT)getStructure()).delete(o));
	else
	    remove();	
    }	    

    /** This method is invoked by GUI in order to delete this visual object. */
    public void delete() {
        if (getParent() instanceof Application) {
            ((Application)getParent()).delete(this);
        } else erase_();
    }
    
    public String toString() {
        return getClassName() + "#" + nbr + " (" + getStructure() + ")";
    }
    
    public String getClassName() {
        String s = ""+getClass();
        return s.substring(s.lastIndexOf(".")+1);
    }

    /**
     * Method for retrieving the VT of the data structure this component
     * represents. By default it is <code>this</code>, but in some cases
     * it can be some other VisualType wrapped in <code>this</code>.
     */
    public VisualType getVisualType() {
	return this;
    }

    /** Translate the location of this component to 
     * the coordinates of the parameter VisualType. 
     */
    public Point getLocation(VisualType vt) {
        if (isShowing()) {
            Point myLoc = getLocationOnScreen();
            Point hisOrigo = vt.getParent().getLocationOnScreen();
            return new Point(myLoc.x - hisOrigo.x, myLoc.y - hisOrigo.y);
        } else {
            Note.err(this, "getLocation(): object not visible " + this);
            return new Point(0,0);
        }
    }

    private StyleSheet defaultStyleSheet = null;
    /** Subclasses should override this method in order to 
     * return defaultStyleSheets for special kinds of entities with
     * unique coloring.
     */
    public StyleSheet getDefaultStyleSheet() {
        if (defaultStyleSheet == null)
            defaultStyleSheet = new StyleSheetAdapter();
        return defaultStyleSheet;
    }

    public void setDefaultStyleSheet(StyleSheet ss) {
	defaultStyleSheet = ss;
    }


    // Sets background color for drawing. 
    void setBackgroundColor(Graphics g) {
  	Color defaultColor;
        ColorDecorator cd;
	if (fdt instanceof ColorDecorator) {
            cd = (ColorDecorator)fdt;
	} else if (fdt instanceof Styled) {
            cd = ((Styled)fdt).getStyleSheet();
	} else 
            cd = getDefaultStyleSheet();
        if (isVisited()) {
            if (hasFocus()) 
                defaultColor = cd.getVisitedFocusedBackgroundColor();
            else
                defaultColor = cd.getVisitedBackgroundColor();
        } else {
            if (hasFocus())
                defaultColor = cd.getFocusedBackgroundColor();
            else
                defaultColor = cd.getDefaultBackgroundColor();
        }
        if (g != null) {
	  //Note.out(this, "Setting backgroundcolor for " + this + " to " + defaultColor +"\n" +
	  //	   "element " + (isVisited() ? "is visited and " : "") + 
	  //(hasFocus() ? "has focus" : "" ));
            g.setColor(defaultColor);
	}
    }
    
    // Sets border color for drawing
    void setBorderColor(Graphics g) {
	Color defaultColor;
        ColorDecorator cd;
	if (fdt instanceof ColorDecorator) {
            cd = (ColorDecorator)fdt;
	} else if (fdt instanceof Styled) {
            cd = ((Styled)fdt).getStyleSheet();
	} else 
            cd = getDefaultStyleSheet();
        if (isVisited()) {
            if (hasFocus()) 
                defaultColor = cd.getVisitedFocusedBorderColor();
            else
                defaultColor = cd.getVisitedBorderColor();
        } else {
            if (hasFocus())
                defaultColor = cd.getFocusedBorderColor();
            else
                defaultColor = cd.getDefaultBorderColor();
        }
        if (g != null)
            g.setColor(defaultColor);
    }

    // Sets pen color for drawing.
    void setPenColor(Graphics g) {
	Color defaultColor;
        ColorDecorator cd;
	if (fdt instanceof ColorDecorator) {
            cd = (ColorDecorator)fdt;
	} else if (fdt instanceof Styled) {
            cd = ((Styled)fdt).getStyleSheet();
	} else 
            cd = getDefaultStyleSheet();
        if (isVisited()) {
            if (hasFocus()) 
                defaultColor = cd.getVisitedFocusedPenColor();
            else
                defaultColor = cd.getVisitedPenColor();
        } else {
            if (hasFocus())
                defaultColor = cd.getFocusedPenColor();
            else
                defaultColor = cd.getDefaultPenColor();
        }
        if (g != null)
            g.setColor(defaultColor); 
    }

    /** This method is invoked when the GUI wants to determine the
     * minimized Shape (Polygon) of this visual object.
     * XXX => VisualComponent?
     * @param x - the x coordinate of the upper left point
     * @param y - the y coordinate of the upper left point
     * @param width - the width of the provided area
     * @param height - the height of the provided area
     */
    public Polygon getMinimizedPolygon(int x, int y, int width, int height) {
        int xpoints[] = new int[4];
        int ypoints[] = new int[4];
        
        xpoints[0] = 0; ypoints[0] = 0;
        xpoints[1] = width; ypoints[1] = 0;
        xpoints[2] = width; ypoints[2] = height;
        xpoints[3] = 0; ypoints[3] = height;
        
        return new Polygon(xpoints, ypoints, 4);
    }
    
    /** This method is invoked when this visual object is printed.
     * Implemented only because, by default, Java prints also
     * invisible components.
     */
    public void print(Graphics g) {
        // bugibugi in Java?
        if (isVisible()) super.print(g);
    }

    public void paint(Graphics g) {
	if (g instanceof MatrixGraphics) {
	    int[] idArray = { this.hashCode(), -1, -1 };
	    if (fdt != null) {
		idArray[1] = fdt.hashCode();
		if (fdt instanceof Key) {
		    Key original = ((Key)fdt).getOrig();
		    if (original != null)
			idArray[2] = original.hashCode();
		}
	    }
	    ((MatrixGraphics)g).setId(idArray);
	    ((MatrixGraphics)g).setName(this.toString());
	}
	// paint frame if this component is dragged
	if (picker.getPick() != null && picked && picker.getPick().equals(this)) {
	    g.drawLine(0,0,3,0);
	    g.drawLine(0,0,0,3);
	    g.drawLine(width - 4, 0, width - 1, 0);
	    g.drawLine(width - 1, 0, width - 1, 3);
	    g.drawLine(width - 4, height - 1, width - 1, height - 1);
	    g.drawLine(width - 1, height - 4, width - 1, height - 1);
	    g.drawLine(0, height - 1, 3, height - 1);
	    g.drawLine(0, height - 1, 0, height - 4);
	    if (width > 20) {
		int center = width/2;
		g.drawLine(center - 2, 0, center + 2, 0);
		g.drawLine(center - 2, height - 1, center + 2, height - 1);
	    }
	    if (height > 20) {
		int center = height/2;
		g.drawLine(0, center - 2, 0, center + 2);
		g.drawLine(width - 1, center - 2 , width - 1, center + 2);
	    }
	}
	super.paint(g);
    }

    private VisualTypeConf conf = null;
    private boolean dragEnabled = true;
    private boolean popUpEnabled = true;
    private boolean activationEnabled = true;

    /** Configure this component and all subcomponents */
    public void configure(VisualTypeConf conf) {
	this.conf = conf;
	if (conf != null) {
	    dragEnabled = conf.isEnabled(this.getClass().getName(), VisualTypeConf.DRAG_OPERATION);
	    popUpEnabled = conf.isEnabled(this.getClass().getName(), VisualTypeConf.POP_UP_MENU_OPERATION);
	    activationEnabled = conf.isEnabled(this.getClass().getName(), VisualTypeConf.HIGHLIGHT_OPERATION);
	}
        for (int i=0; i<getItemCount(); i++)
            getItem(i).configure(conf);
        for (int i=0; i<getLinkCount(); i++)
            getLink(i).configure(conf);
	for (int i=0; i<getOtherCount(); i++)
	    if (getOther(i) instanceof VisualType)
		((VisualType)getOther(i)).configure(conf);
    }

    /** Is drag enabled for this object */
    protected boolean dragEnabled() {
	if (conf != null)
	    return dragEnabled;
	else
	    return simulator;
    }

    /** Is drop enabled for this object */
    protected boolean dropEnabled() {
	if (entered == null)
	    return simulator;
	VisualTypeConf entconf = entered.getConf();
	if (entconf != null)
	    return entconf.isEnabled(entered.getClass().getName(), VisualTypeConf.DROP_OPERATION);
	else
	    return simulator;
    }

    /** Is pop-up menu enabled for this object */
    protected boolean popUpEnabled() {
	if (conf != null)
	    return popUpEnabled;
	else
	    return simulator;
    }

    /** Is activation enabled for this object */
    protected boolean activationEnabled() {
	if (conf != null) 
	    return activationEnabled;
	else 
	    return simulator;
    }

    /** Return the VisualTypeConf of this object. */
    protected VisualTypeConf getConf() {
	if (conf == null) 
	    if (getParent() instanceof VisualType)
		return ((VisualType)getParent()).getConf();
	    else
		return null;
	else
	    return conf;
    }


    /**
     * Changes the name of the current representation.
     *
     * @param name of the representation
     */
    public void setNameOfRepresentation(String name) {
	nameOfRepresentation = name;
    }

    /**
     * Returns the name of the current representation.
     *
     * @return the name of the current representation
     */
    public String getNameOfRepresentation() {
	return nameOfRepresentation;
    }

    protected void validateTree() {
	doLayout();
    }

    /**
     * Returns all visualizations that visualize the given
     * structure.
     */
    public static VisualType[] getVisualizations(FDT structure) {
	Vector vis = (Vector)visualizations.get(structure);
	if (vis == null)
	    return new VisualType[0];
	VisualType[] visuals = new VisualType[vis.size()];
	vis.copyInto(visuals);
	return visuals;
    }

    /**
     * Change representation of child item.
     * @param oldType Item to replace with a new representation.
     * @param newType Name of new representation.
     */
    public void changeRepresentation(VisualContainer oldType, String newType) {
	int index;
	for(index=0;index<getComponentCount();index++) {	    
	    if (getComponent(index)==oldType)
		break;
	}
	if (index>=getComponentCount()) {
            Note.err(this, "Item not in parent when changing representation.");
	    return;
	}

	FDT s = oldType.getStructure();
	remove(index);
	String name = oldType.getName();
	VisualType vc =
	    RepresentationFactory.getRepresentation(s, newType);
	vc.setNameOfRepresentation(newType);
	add(vc, index);
	vc.rename(name); /* Rename handles revalidation. */
    }
}
